"""
Draws a frame on a GtkDrawingArea. This frame may or may not be resizable.
This is usually used for the scanner calibration dialog or the page image
cropping.
"""

import logging
import math

try:
    import gi
    gi.require_version('Gdk', '3.0')
    from gi.repository import Gdk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Drawer(object):
    HANDLE_RADIUS = 10
    HANDLE_DEFAULT_COLOR = (0.0, 0.25, 1.0, 0.5)
    HANDLE_HOVER_COLOR = (0.1, 0.75, 0.1, 1.0)
    HANDLE_SELECTED_COLOR = (0.75, 0.1, 0.1, 1.0)
    HANDLE_MIN_DIST = 50

    RECTANGLE_COLOR = (0.0, 0.25, 1.0, 1.0)
    OUTTER_COLOR = (0.0, 0.25, 1.0, 0.25)

    FRAME_CORNERS = [
        (0, 1),
        (0, 3),
        (2, 1),
        (2, 3),
    ]

    def __init__(
            self, core, drawing_area, content_full_size,
            get_frame_cb, set_frame_cb=None):
        """
        Arguments:
        - drawing_area: the drawing area on which the frame must be
          represented
        - content_full_size: the drawing area has its own size, smaller than
          the actual scan or image. We need to know the actual size of the
          content, so we can figure out at which scale the content and the
          frame are to represented.
        - get_frame_cb : callback returning the frame to display relative
          to content_full_size. A callback is provided so we can always have
          an up-to-date value (for instance coming from the configuration).
        """
        self.core = core

        if drawing_area.get_window() is None:
            drawing_area.connect("realize", self.on_realize)
        else:
            self.on_realize(drawing_area)

        self.drawing_area = drawing_area
        self.get_frame_cb = get_frame_cb
        self.set_frame_cb = set_frame_cb
        self.content_full_size = content_full_size
        self.active_handle = None
        self.selected_handle = None

        self.draw_connect_id = drawing_area.connect("draw", self.on_draw)

        self.motion_connect_id = None
        self.press_connect_id = None
        self.release_connect_id = None
        if self.set_frame_cb:
            self.motion_connect_id = drawing_area.connect(
                "motion-notify-event", self.on_motion
            )
            self.press_connect_id = drawing_area.connect(
                "button-press-event", self.on_pressed
            )
            self.release_connect_id = drawing_area.connect(
                "button-release-event", self.on_released
            )

    def stop(self):
        if self.draw_connect_id is not None:
            self.drawing_area.disconnect(self.draw_connect_id)
        if self.motion_connect_id is not None:
            self.drawing_area.disconnect(self.motion_connect_id)
        if self.press_connect_id is not None:
            self.drawing_area.disconnect(self.press_connect_id)
        if self.release_connect_id is not None:
            self.drawing_area.disconnect(self.release_connect_id)
        self.draw_connect_id = None
        self.motion_connect_id = None
        self.press_connect_id = None
        self.release_connect_id = None

    def request_redraw(self):
        self.drawing_area.queue_draw()

    def _get_factor(self):
        widget_height = self.drawing_area.get_allocated_height()
        widget_width = self.drawing_area.get_allocated_width()
        factor_w = self.content_full_size[0] / widget_width
        factor_h = self.content_full_size[1] / widget_height
        factor = max(factor_w, factor_h)
        return factor

    def _get_frame(self):
        factor = self._get_factor()
        frame = self.get_frame_cb()
        if frame is None:
            frame = (
                0, 0, self.content_full_size[0], self.content_full_size[1]
            )
        return (
            frame[0] / factor, frame[1] / factor,
            frame[2] / factor, frame[3] / factor,
        )

    def _get_closest_handle(self, x, y):
        frame = self._get_frame()

        handle_dists = [
            (
                math.hypot(frame[corner_x] - x, frame[corner_y] - y),
                (corner_x, corner_y)
            )
            for (corner_x, corner_y) in self.FRAME_CORNERS
        ]
        handle = min(handle_dists)
        if handle[0] > self.HANDLE_MIN_DIST:
            return None
        return handle[1]

    def on_realize(self, drawing_area, *args, **kwargs):
        mask = (
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK
        )
        drawing_area.add_events(mask)
        drawing_area.get_window().set_events(
            drawing_area.get_window().get_events() | mask
        )

    def on_motion(self, drawing_area, event):
        if self.selected_handle is None:
            active_handle = self._get_closest_handle(event.x, event.y)
            if active_handle != self.active_handle:
                self.request_redraw()
            self.active_handle = active_handle
        else:
            assert self.set_frame_cb is not None
            factor = self._get_factor()
            frame = self.get_frame_cb()
            if frame is not None:
                frame = list(frame)
            else:
                frame = [
                    0, 0, self.content_full_size[0], self.content_full_size[1]
                ]
            x = event.x * factor
            y = event.y * factor
            frame[self.selected_handle[0]] = x
            frame[self.selected_handle[1]] = y
            frame = (
                max(0, min(frame[0], frame[2])),
                max(0, min(frame[1], frame[3])),
                min(self.content_full_size[0], max(frame[0], frame[2])),
                min(self.content_full_size[1], max(frame[1], frame[3])),
            )
            self.set_frame_cb(frame)
            self.request_redraw()

    def on_pressed(self, drawing_area, event):
        self.selected_handle = self._get_closest_handle(event.x, event.y)
        self.request_redraw()

    def on_released(self, drawing_area, event):
        self.selected_handle = None
        self.request_redraw()

    def on_draw(self, drawing_area, cairo_ctx):
        frame = self._get_frame()
        widget_height = self.drawing_area.get_allocated_height()
        widget_width = self.drawing_area.get_allocated_width()

        # outter
        cairo_ctx.save()
        try:
            color = self.OUTTER_COLOR
            cairo_ctx.set_source_rgba(color[0], color[1], color[2], color[3])

            outters = [
                (0, 0, widget_width, frame[1]),
                (0, frame[3], widget_width, widget_height - frame[3]),
                (0, frame[1], frame[0], frame[3] - frame[1]),
                (
                    frame[2], frame[1],
                    widget_width - frame[2], frame[3] - frame[1]
                ),
            ]
            for (x, y, w, h) in outters:
                cairo_ctx.rectangle(x, y, w, h)

            cairo_ctx.fill()
        finally:
            cairo_ctx.restore()

        (x, y, w, h) = (
            frame[0], frame[1],
            frame[2] - frame[0], frame[3] - frame[1]
        )

        # rectangle
        cairo_ctx.save()
        try:
            color = self.RECTANGLE_COLOR
            cairo_ctx.set_source_rgba(color[0], color[1], color[2], color[3])
            cairo_ctx.set_line_width(1.0)
            cairo_ctx.rectangle(x, y, w, h)
            cairo_ctx.stroke()
        finally:
            cairo_ctx.restore()

        if self.set_frame_cb is None:
            return

        # handles
        cairo_ctx.save()
        try:
            for corner in self.FRAME_CORNERS:
                if self.selected_handle == corner:
                    color = self.HANDLE_SELECTED_COLOR
                elif self.active_handle == corner:
                    color = self.HANDLE_HOVER_COLOR
                else:
                    color = self.HANDLE_DEFAULT_COLOR
                cairo_ctx.set_source_rgba(
                    color[0], color[1], color[2], color[3]
                )
                cairo_ctx.arc(
                    frame[corner[0]], frame[corner[1]],
                    self.HANDLE_RADIUS, 0., 2 * math.pi
                )
                cairo_ctx.fill()
        finally:
            cairo_ctx.restore()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        # drawing area --> Drawer
        self.active_drawers = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_drawer_frame',
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def draw_frame_start(
            self, drawing_area, content_full_size,
            get_frame_cb, set_frame_cb=None):
        drawer = Drawer(
            self.core, drawing_area, content_full_size,
            get_frame_cb, set_frame_cb
        )
        self.active_drawers[drawing_area] = drawer
        return drawer

    def draw_frame_stop(self, drawing_area):
        if drawing_area not in self.active_drawers:
            return
        drawer = self.active_drawers.pop(drawing_area)
        drawer.stop()
        return drawer


if __name__ == "__main__":
    import sys
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk

    core = openpaperwork_core.Core()
    core._load_module("test", sys.modules[__name__])
    core.init()

    window = Gtk.Window()
    window.set_size_request(600, 600)

    box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 20)
    window.add(box)

    drawing_area_a = Gtk.DrawingArea()
    box.pack_start(drawing_area_a, expand=True, fill=True, padding=0)
    drawing_area_b = Gtk.DrawingArea()
    box.pack_start(drawing_area_b, expand=True, fill=True, padding=0)

    frame = (30, 50, 100, 110)

    def get_frame():
        return frame

    def set_frame(f):
        global frame
        frame = f
        drawing_area_a.queue_draw()
        drawing_area_b.queue_draw()

    core.call_success(
        "draw_frame_start", drawing_area_a, (200, 300), get_frame
    )
    core.call_success(
        "draw_frame_start", drawing_area_b, (200, 300), get_frame, set_frame
    )

    window.show_all()

    Gtk.main()



import logging

try:
    import gi
    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = (None, None)
        self.updated = set()
        self.promise = None

    def get_interfaces(self):
        return [
            'doc_open',
            'gtk_drag_and_drop'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.promise = openpaperwork_core.promise.Promise(self.core)

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def _parse_paperwork_uri(self, uri):
        infos = uri.split("#", 1)[1]
        infos = infos.split("&")
        infos = (i.split("=", 1) for i in infos)
        infos = {k: v for (k, v) in infos}
        return (infos['doc_id'], int(infos['page']))

    def doc_close(self):
        self.active_doc = (None, None)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def drag_and_drop_page_enable(self, widget):
        """
        Arguments:
            widget - GTK widget on which page drag'n'drop must be enabled.

        When a drop is received, it will call the method
        'drag_and_drop_get_destination(widget, x, y)'.
        The plugin to whom the widget belongs should reply with
        (doc_id, page_idx).
        """
        widget.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.MOVE)
        targets = Gtk.TargetList.new([])
        targets.add_uri_targets(0)
        widget.drag_dest_set_target_list(targets)
        widget.connect("drag-data-received", self._on_drag_data_received)

    def _on_drag_data_received(
            self, widget, drag_context, x, y, selection_data, info, time):
        uris = selection_data.get_uris()
        LOGGER.info(
            "drag_data_received(%s, %d, %d, %s, %s)",
            widget, x, y, uris, info
        )

        dst = self.core.call_success(
            "drag_and_drop_get_destination", widget, x, y
        )
        if dst is None:
            LOGGER.error("Nobody accepted a drop on %s (%d, %d)", widget, x, y)
            return

        (dst_doc_id, dst_doc_url, dst_page_idx) = dst

        for uri in reversed(uris):
            LOGGER.info("Drop: URI: %s", uri)

            self.core.call_all(
                "drag_and_drop_page_add",
                uri, dst_doc_id, dst_doc_url, dst_page_idx
            )

        self.core.call_all("drag_and_drop_apply")

    def drag_and_drop_page_add(
            self, src_uri, dst_doc_id, dst_doc_url, dst_page_idx):
        LOGGER.info("Drop: %s --> %s p%d", src_uri, dst_doc_id, dst_page_idx)

        if "doc_id=" not in src_uri or "page=" not in src_uri:
            LOGGER.info("Drop: Import of %s", src_uri)
            # TODO(Jflesch): Should import to the target document (drop)
            # instead of the currently-opened document
            self.core.call_all("gtk_doc_import", [src_uri])
            return
        else:
            # moving page inside the current document
            (src_doc_id, src_page_idx) = self._parse_paperwork_uri(src_uri)

        src_doc_url = self.core.call_success("doc_id_to_url", src_doc_id)
        assert src_doc_url is not None

        if src_doc_id == dst_doc_id and src_page_idx < dst_page_idx:
            dst_page_idx -= 1
        dst_page_idx = max(dst_page_idx, 0)

        LOGGER.info(
            "Drop: %s p%d --> %s p%d",
            src_doc_id, src_page_idx,
            dst_doc_id, dst_page_idx
        )

        if src_doc_id == dst_doc_id and src_page_idx == dst_page_idx:
            return

        self.promise = self.promise.then(
            self.core.call_all, "page_move_by_url",
            src_doc_url, src_page_idx,
            dst_doc_url, dst_page_idx
        )
        self.promise = self.promise.then(lambda *args, **kwargs: None)

        for doc in ((src_doc_id, src_doc_url), (dst_doc_id, dst_doc_url)):
            self.promise = self.promise.then(
                self.core.call_all, "doc_reload", *doc
            )
            self.promise = self.promise.then(lambda *args, **kwargs: None)

        self.updated.add(src_doc_id)
        self.updated.add(dst_doc_id)

    def drag_and_drop_apply(self):
        if len(self.updated) <= 0:
            return

        self.promise = self.promise.then(self.core.call_success(
            "transaction_simple_promise",
            [('upd', doc_id) for doc_id in self.updated]
        ))
        self.core.call_success("transaction_schedule", self.promise)

        self.updated = set()
        self.promise = openpaperwork_core.promise.Promise(self.core)



import collections
import distro
import logging
import os
import sys

from . import util
from .. import (_, PluginBase)


LOGGER = logging.getLogger(__name__)


PACKAGE_TOOLS = {
    'debian': 'apt-get install -y',
    'fedora': 'dnf install',
    'gentoo': 'emerge',
    'linuxmint': 'apt-get install -y',
    'raspbian': 'apt-get install -y',
    'suse': 'zypper in',
    'ubuntu': 'apt-get install -y',
}


class Plugin(PluginBase):
    def __init__(self):
        self.console = None

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        # will call method from interface 'chkdeps', but we can still
        # work if no other plugin implement 'chkdeps'.
        return []

    def _get_distribution(self):
        distribution = distro.linux_distribution(full_distribution_name=False)
        self.console.print(f"Detected system: {' '.join(distribution)}")
        distribution = distribution[0].lower()
        if distribution not in PACKAGE_TOOLS:
            LOGGER.warning(
                "WARNING: Unknown distribution."
                " Can't suggest packages to install"
            )
        return distribution

    def cmd_set_console(self, console):
        self.console = console

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'chkdeps',
            help=_("Check that all required dependencies are installed")
        )
        p.add_argument(
            "--yes", "-y",
            required=False, default=False, action='store_true'
        )

    def cmd_run(self, console, args):
        if args.command != 'chkdeps':
            return None
        auto = args.yes

        if auto:
            LOGGER.warning("Confirmation disabled")

        distribution = self._get_distribution()

        missing = collections.defaultdict(dict)
        self.core.call_all("chkdeps", missing)

        if len(missing) > 0:
            console.print(_("Missing dependencies:"))
            for (dep_name, distrib_packages) in missing.items():
                console.print(_("- {dep_name} (package: {pkg_name})").format(
                    dep_name=dep_name,
                    pkg_name=(
                        distrib_packages[distribution]
                        if distribution in distrib_packages
                        else _("UNKNOWN")
                    )
                ))

        if distribution in PACKAGE_TOOLS:
            command = PACKAGE_TOOLS[distribution]
            if os.getuid() != 0:
                command = "sudo {}".format(command)
        else:
            command = None
        has_pkg = False
        for (dep_name, distrib_packages) in missing.items():
            if distribution in distrib_packages:
                if command is not None:
                    command += " " + distrib_packages[distribution]
                has_pkg = True

        if has_pkg and command is not None:
            console.print("")
            console.print(_("Suggested command:"))
            console.print("  " + command)
            console.print("")
            if not auto:
                r = util.ask_confirmation(
                    console,
                    _("Do you want to run this command now ?"),
                    default_interactive='n',
                    default_non_interactive='n',
                )
                if r != 'y':
                    return {
                        "missing": missing,
                        "command": command,
                    }
            console.print("Running command ...")
            r = os.system(command)
            console.print("Command returned {}".format(r))
            if r != 0:
                sys.exit(r)
        elif len(missing) > 0:
            console.print(
                _("Don't know how to install missing dependencies. Sorry.")
            )
        else:
            console.print(_("Nothing to do."))

        return {
            "missing": missing,
            "command": command,
        }



import logging

from . import (
    AbstractExportPipe,
    AbstractExportPipePlugin,
    ExportData,
    ExportDataType
)
from .. import _


LOGGER = logging.getLogger(__name__)


class DocSetToDoc(AbstractExportPipe):
    def __init__(self, core):
        super().__init__(
            name="doc_set_to_docs",
            input_type=ExportDataType.DOCUMENT_SET,
            output_type=ExportDataType.DOCUMENT
        )
        self.core = core

    def export(self, input_data, result='final', target_file_url=None):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET
        assert not input_data.expanded

        docs = input_data.data
        if result == 'preview':
            docs = docs[:1]
        children = [
            ExportData(ExportDataType.DOCUMENT, doc)
            for doc in docs
        ]
        input_data.set_children(children)
        return input_data

    def get_estimated_size_factor(self, input_data):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET
        return len(input_data.data)

    def __str__(self):
        # this pipe shouldn't ever be visible to end-user
        return "Expand document set into documents (internal)"


class DocToPage(AbstractExportPipe):
    def __init__(self, core):
        super().__init__(
            name="doc_to_pages",
            input_type=ExportDataType.DOCUMENT,
            output_type=ExportDataType.PAGE
        )
        self.core = core

    def can_export_doc(self, doc_url):
        return True

    def export(self, input_data, result='final', target_file_url=None):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET

        docs = input_data.iter(ExportDataType.DOCUMENT)
        for (doc_set, doc) in docs:
            assert not doc.expanded
            nb_pages = self.core.call_success(
                "doc_get_nb_pages_by_url", doc.data[1]
            )
            pages = [
                ExportData(ExportDataType.PAGE, page_idx)
                for page_idx in range(0, nb_pages)
            ]
            doc.set_children(pages)

        return input_data

    def get_estimated_size_factor(self, input_data):
        # average number of pages per document
        assert input_data.dtype == ExportDataType.DOCUMENT_SET
        nb_pages = 0
        nb_docs = len(input_data.data)
        for (doc_id, doc_url) in input_data.data:
            nb_pages += self.core.call_success(
                "doc_get_nb_pages_by_url", doc_url
            )
        return nb_pages / nb_docs

    def __str__(self):
        return _("Page by page processing")


class Plugin(AbstractExportPipePlugin):
    def get_deps(self):
        return [
            {
                'interface': 'pages',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
        ]

    def get_interfaces(self):
        return super().get_interfaces() + [
            "export_pipes_generic",
        ]

    def init(self, core):
        super().init(core)
        self.pipes = [
            DocSetToDoc(self.core),
            DocToPage(self.core),
        ]



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2014  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>

import logging
import os


LOGGER = logging.getLogger(__name__)


def rm_rf(path):
    """
    Act as 'rm -rf' in the shell
    """
    if os.path.isfile(path):
        os.unlink(path)
    elif os.path.isdir(path):
        for root, dirs, files in os.walk(path, topdown=False):
            for filename in files:
                filepath = os.path.join(root, filename)
                LOGGER.info("Deleting file %s" % filepath)
                os.unlink(filepath)
            for dirname in dirs:
                dirpath = os.path.join(root, dirname)
                if os.path.islink(dirpath):
                    LOGGER.info("Deleting link %s" % dirpath)
                    os.unlink(dirpath)
                else:
                    LOGGER.info("Deleting dir %s" % dirpath)
                    os.rmdir(dirpath)
        LOGGER.info("Deleting dir %s", path)
        os.rmdir(path)


def levenshtein_distance(
            str_a: str, str_b: str,
            str_a_idx: int = 0, str_b_idx: int = 0
        ):
    if str_a_idx == len(str_a) or str_b_idx == len(str_b):
        return len(str_a) - str_a_idx + len(str_b) - str_b_idx

    # no change required
    if str_a[str_a_idx] == str_b[str_b_idx]:
        return levenshtein_distance(str_a, str_b, str_a_idx + 1, str_b_idx + 1)

    return 1 + min(
        # insert character
        levenshtein_distance(str_a, str_b, str_a_idx, str_b_idx + 1),
        # delete character
        levenshtein_distance(str_a, str_b, str_a_idx + 1, str_b_idx),
        # replace character
        levenshtein_distance(str_a, str_b, str_a_idx + 1, str_b_idx + 1),
    )



import logging

try:
    from gi.repository import Gio
    GIO_AVAILABLE = True
except (ImportError, ValueError):
    GIO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1000

    def __init__(self):
        super().__init__()
        self.menu = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'menu',
            'menu_app',
            'menu_app_help',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_app_help',
                'defaults': ['paperwork_gtk.actions.app.help'],
            },
            {
                'interface': 'help_documents',
                'defaults': ['paperwork_gtk.model.help'],
            },
            {
                'interface': 'gtk_app_menu',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GIO_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def init(self, core):
        super().init(core)
        if not GIO_AVAILABLE:
            return
        self.menu = Gio.Menu()

    def on_doclist_initialized(self):
        for (title, file_name) in self.core.call_success("help_get_files"):
            item = Gio.MenuItem.new(title, "win.open_help." + file_name)
            self.menu.append_item(item)
        self.core.call_all("menu_app_append_submenu", _("Help"), self.menu)

    def help_add_menu_item(self, menu_item):
        self.menu.append_item(menu_item)



import logging

from . import (
    AbstractExportPipe,
    AbstractExportPipePlugin,
    AbstractSimpleTransformExportPipe,
    ExportData,
    ExportDataType
)
from .. import _


LOGGER = logging.getLogger(__name__)


class ExportDataPage(ExportData):
    """
    Page images takes a lot of memory --> we only load them when actually
    requested.
    """
    def __init__(self, core, doc_id, doc_url, page_idx, progress, final=False):
        super().__init__(ExportDataType.PAGE, page_idx)
        self.expanded = True

        self.core = core
        self.doc_id = doc_id
        self.doc_url = doc_url
        self.page_idx = page_idx
        self.progress = progress
        self.final = final

    def get_children(self):
        # ASSSUMPTION(Jflesch): pages are always requested sequentially
        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "on_progress", "export", self.progress,
            _("Exporting {doc_id} p{page_idx} ...").format(
                doc_id=self.doc_id, page_idx=(self.page_idx + 1)
            )
        )

        page_url = (self.core.call_success(
            "page_get_img_url", self.doc_url, self.page_idx
        ))
        assert page_url is not None
        img = self.core.call_success("url_to_pillow", page_url)
        if img is None:
            # WORKAROUND(Jflesch): may happen with corrupted PDFs
            return
        boxes = self.core.call_success(
            "page_get_boxes_by_url", self.doc_url, self.page_idx
        ) or []

        yield ExportData(ExportDataType.IMG_BOXES, (img, boxes))

        if self.final:
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_progress", "export", 1.0
            )


class DocToPillowBoxesExportPipe(AbstractExportPipe):
    def __init__(self, core):
        super().__init__(
            name="img_boxes",
            input_type=ExportDataType.PAGE,
            output_type=ExportDataType.IMG_BOXES
        )
        self.core = core

    def can_export_page(self, doc_url, page_nb):
        return True

    def export(self, input_data, result='final', target_file_url=None):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET

        docs = input_data.iter(ExportDataType.DOCUMENT)
        docs = list(docs)

        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "on_progress", "export", 0.0,
            _("Exporting ...")
        )

        total_pages = 0
        for (doc_set, doc) in docs:
            assert doc.expanded
            total_pages += len(doc.get_children())

        nb_pages = 0
        last_page = None
        for (doc_set, doc) in docs:
            assert doc.expanded

            # replace the document page list by objects that will
            # generate their children (img+boxes) on-the-fly.
            new_pages = [
                ExportDataPage(
                    self.core, doc.data[0], doc.data[1], page.data,
                    (nb_pages + idx) / total_pages
                )
                for (idx, page) in enumerate(doc.get_children())
            ]
            doc.set_children(new_pages)
            nb_pages += len(new_pages)
            last_page = new_pages[-1]
        last_page.final = True

        return input_data

    def __str__(self):
        return _("Split page(s) into image(s) and text(s)")


class PageToImageExportPipe(AbstractExportPipe):
    def __init__(self, core, name, format, file_extensions, mime, has_quality):
        super().__init__(
            name=name,
            input_type=ExportDataType.IMG_BOXES,
            output_type=ExportDataType.OUTPUT_URL_FILE,
        )
        self.can_change_quality = has_quality
        self.format = format
        self.file_extensions = file_extensions
        self.mime = mime
        self.core = core

    def export(self, input_data, result='final', target_file_url=None):
        if target_file_url is None:
            (target_file_url, file_desc) = self.core.call_success(
                "fs_mktemp", prefix="paperwork-export-",
                suffix="." + self.file_extensions[0], mode="w"
            )
            file_desc.close()

        list_img_boxes = input_data.iter(ExportDataType.IMG_BOXES)

        out_files = []
        for (doc_set, (doc, (page, img_boxes))) in list_img_boxes:
            out = target_file_url
            if len(out_files) != 0:
                out = target_file_url.rsplit(".", 1)
                out = "{}_{}.{}".format(out[0], len(out_files), out[1])

            self.core.call_success(
                "pillow_to_url", img_boxes.data[0], out,
                format=self.format, quality=self.quality
            )
            out_files.append(out)
        return out_files

    def get_output_mime(self):
        return (self.mime, self.file_extensions)

    def __str__(self):
        return _("Image file ({})".format(self.format))


class BlackAndWhiteExportPipe(AbstractSimpleTransformExportPipe):
    def __init__(self, core):
        super().__init__(core, "bw")

    def transform(self, pil_img):
        pil_img = pil_img.convert("L")  # to grayscale
        pil_img = pil_img.point(lambda x: 0 if x < 128 else 255, '1')
        return pil_img

    def __str__(self):
        return _("Black & White")


class GrayscaleExportPipe(AbstractSimpleTransformExportPipe):
    def __init__(self, core):
        super().__init__(core, "grayscale")

    def transform(self, pil_img):
        return pil_img.convert("L")

    def __str__(self):
        return _("Grayscale")


class Plugin(AbstractExportPipePlugin):
    def init(self, core):
        super().init(core)
        self.pipes = [
            BlackAndWhiteExportPipe(core),
            DocToPillowBoxesExportPipe(core),
            GrayscaleExportPipe(core),
            PageToImageExportPipe(
                core, "bmp", "BMP", ("bmp",), "image/x-ms.bmp", False
            ),
            PageToImageExportPipe(
                core, "gif", "GIF", ("gif",), "image/gif", False
            ),
            PageToImageExportPipe(
                core, "jpeg", "JPEG", ("jpeg", "jpg"), "image/jpeg", True
            ),
            PageToImageExportPipe(
                core, "png", "PNG", ("png",), "image/png", False
            ),
            PageToImageExportPipe(
                core, "tiff", "TIFF", ("tiff", "tif"), "image/tiff", False
            ),
        ]

    def get_interfaces(self):
        return super().get_interfaces() + [
            "export_pipes_img",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_text',
                'defaults': [
                    # load also model.img because model.pdf will
                    # satisfy the interface 'page_img' anyway
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'page_img',
                'defaults': [
                    # load also model.hocr because model.pdf will
                    # satisfy the interface 'doc_text' anyway
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
        ]



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000000

    def __init__(self):
        super().__init__()
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_settings',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.workdir = self.core.call_success("config_get", "workdir")

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _show_workdir_info(self, *args, **kwargs):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings", "storage.glade"
        )
        dialog = widget_tree.get_object("workdir_info_dialog")
        dialog.set_transient_for(self.windows[-1])
        dialog.set_modal(True)
        dialog.connect("response", self._on_workdir_info_dialog_closed)
        dialog.connect("destroy", self._on_workdir_info_dialog_closed)
        dialog.set_visible(True)

    def _on_workdir_info_dialog_closed(self, dialog, *args, **kwargs):
        dialog.set_visible(False)

    def complete_settings(self, global_widget_tree):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings", "storage.glade"
        )

        workdir_button = widget_tree.get_object("work_dir_chooser_button")
        basename = self.core.call_success("fs_basename", self.workdir)
        workdir_button.set_label(basename)
        workdir_button.connect("clicked", self._on_button_clicked, widget_tree)

        workdir_info_button = widget_tree.get_object("button_workdir_info")
        workdir_info_button.connect("clicked", self._show_workdir_info)

        self.core.call_success(
            "add_setting_to_dialog", global_widget_tree, _("Storage"),
            [widget_tree.get_object("workdir")],
            extra_widget=workdir_info_button,
        )

    def _on_button_clicked(self, button, widget_tree):
        dialog = Gtk.FileChooserDialog(
            _("Work Directory"),
            self.windows[-1],
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT
            )
        )
        dialog.set_modal(True)
        dialog.set_local_only(False)
        workdir = self.core.call_success("config_get", "workdir")
        if self.core.call_success("fs_exists", workdir):
            dialog.set_uri(workdir)
        dialog.connect("response", self._on_file_dialog_response, widget_tree)
        dialog.show_all()

    def _on_file_dialog_response(self, dialog, response_id, widget_tree):
        if (response_id != Gtk.ResponseType.ACCEPT and
                response_id != Gtk.ResponseType.OK and
                response_id != Gtk.ResponseType.YES and
                response_id != Gtk.ResponseType.APPLY):
            LOGGER.info("User canceled (response_id=%d)", response_id)
            dialog.destroy()
            return

        workdir = dialog.get_uri()
        dialog.set_visible(False)

        LOGGER.info("Setting work directory to %s", workdir)
        self.core.call_all("config_put", "workdir", workdir)

        # Bug report 170: Make sure the current document (in the old work
        # directory) is closed so the user cannot use it by accident anymore
        self.core.call_all("doc_close")

        basename = self.core.call_success("fs_basename", workdir)
        widget_tree.get_object("work_dir_chooser_button").set_label(basename)

        # Basic check of work directory content to avoid common mistake.
        # We do not prevent the user from selecting the folder. We just
        # warn them.
        workdir_content = self.core.call_success("fs_listdir", workdir)
        show_info_dialog = False
        for file_url in workdir_content:
            file_name = self.core.call_success("fs_basename", file_url)
            if file_name[0] == ".":
                continue
            if "." in file_name:
                # assume it's a file ; anyway, it doesn't have the usual
                # name format that Paperwork folders have
                LOGGER.warning(
                    "Suspect file name found in work directory: %s",
                    file_name
                )
                show_info_dialog = True
                break
        if show_info_dialog:
            self._show_workdir_info()

    def config_save(self):
        workdir = self.core.call_success("config_get", "workdir")

        if workdir != self.workdir:
            LOGGER.info("Work directory has been changed --> Synchronizing")
            self.core.call_all("transaction_sync_all")

        self.workdir = workdir



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_windows = []
        self.sections = {}
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_settings_dialog',
            'gtk_window_listener',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_app_menu',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.active_windows.append(window)

    def on_gtk_window_closed(self, window):
        self.active_windows.remove(window)

    def gtk_open_settings(self, *args, **kwargs):
        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.settings", "settings.css"
        )

        global_widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings", "settings.glade"
        )
        self.widget_tree = global_widget_tree
        self.core.call_all('complete_settings', global_widget_tree)
        settings = global_widget_tree.get_object("settings_window")
        settings.set_transient_for(self.active_windows[-1])
        settings.set_modal(True)
        settings.connect("destroy", self._save_settings, global_widget_tree)
        settings.set_visible(True)
        self.core.call_all("on_gtk_window_opened", settings)

    def close_settings(self):
        if self.widget_tree is not None:
            dialog = self.widget_tree.get_object("settings_window")
            dialog.set_visible(False)
            self.widget_tree = None

    def on_quit(self):
        self.close_settings()

    def _save_settings(self, window, global_widget_tree):
        LOGGER.info("Settings closed. Saving configuration")
        self.core.call_all("config_save")
        self.core.call_all("on_gtk_window_closed", window)
        self.core.call_all("on_settings_closed", global_widget_tree)
        self.widget_tree = None

    def add_setting_to_dialog(
            self, global_widget_tree, title, widgets, extra_widget=None):
        """
        Add a setting or a set of settings to the main screen in the settings
        dialog.
        """
        # We have many setting boxes to add to the settings box.
        # --> we need many copies of the setting box --> we load many times
        # the widget tree
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings", "settings_section.glade"
        )
        widget_tree.get_object("setting_section_name").set_text(title)

        if extra_widget:
            box = widget_tree.get_object("settings_title_box")
            box.pack_start(extra_widget, expand=False, fill=False, padding=0)

        inner = widget_tree.get_object("setting_box")
        for widget in widgets:
            inner.pack_start(widget, expand=False, fill=True, padding=0)
        global_widget_tree.get_object("settings_box").pack_start(
            widget_tree.get_object("setting_section"),
            expand=False, fill=True, padding=0
        )
        self.sections[title] = widget_tree.get_object("setting_section")
        return True

    def add_setting_screen(
            self, global_widget_tree, name, widget_header, widget_body):
        global_widget_tree.get_object("settings_stack_header").add_named(
            widget_header, name
        )
        global_widget_tree.get_object("settings_stack_body").add_named(
            widget_body, name
        )

    def show_setting_screen(self, global_widget_tree, name):
        global_widget_tree.get_object(
            "settings_stack_header"
        ).set_visible_child_name(name)
        global_widget_tree.get_object(
            "settings_stack_body"
        ).set_visible_child_name(name)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("settings_window"),
            self.core.call_success("fs_join", out_dir, "settings.png")
        )
        for (name, section) in self.sections.items():
            name = name.lower().replace(" ", "_")
            self.core.call_success(
                "screenshot_snap_widget", section,
                self.core.call_success(
                    "fs_join", out_dir, "settings_{}.png".format(name)
                ),
                margins=(100, 100, 100, 100)
            )

    def settings_scroll_to_top(self):
        scroll = self.widget_tree.get_object("settings_scrolled_window")
        vadj = scroll.get_vadjustment()
        vadj.set_value(vadj.get_lower())

    def settings_scroll_to_bottom(self):
        scroll = self.widget_tree.get_object("settings_scrolled_window")
        vadj = scroll.get_vadjustment()
        vadj.set_value(vadj.get_upper())



"""
Ensure calibration is displayed when a scan is running.
"""

import logging

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Drawer(object):
    def __init__(self, core, drawing_area, read_only=False):
        self.core = core
        self.drawing_area = drawing_area
        self.content_full_size = None
        self.scan_id = None
        self.active = False
        self.drawer = None
        self.read_only = read_only

    def _get_frame(self):
        return self.core.call_success("config_get", "scanner_calibration")

    def _set_frame(self, frame):
        self.core.call_all("config_put", "scanner_calibration", frame)

    def set_content_full_size(self, size):
        self.content_full_size = size

    def start(self):
        if self.active:
            self.stop()

        self.core.call_all(
            "config_add_observer", "scanner_calibration",
            self.request_redraw
        )
        self.drawer = self.core.call_success(
            "draw_frame_start", self.drawing_area, self.content_full_size,
            self._get_frame,
            self._set_frame if not self.read_only else None
        )
        self.active = True

    def request_redraw(self, *args, **kwargs):
        if self.drawer is None:
            return
        self.drawer.request_redraw()

    def stop(self):
        if not self.active:
            return
        self.core.call_all("draw_frame_stop", self.drawing_area)
        self.core.call_all(
            "config_remove_observer", "scanner_calibration",
            self.request_redraw
        )
        self.active = False


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.drawers = {}  # drawing_area --> drawer
        self.scan_id_to_drawers = {}  # int --> drawer

    def get_interfaces(self):
        return [
            'gtk_drawer_calibration',
            'gtk_drawer_scan',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_drawer_frame',
                'defaults': ['paperwork_gtk.drawer.frame'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
            {
                'interface': 'scanner_calibration',
                'defaults': [
                    'paperwork_backend.guesswork.cropping.calibration'
                ]
            },
        ]

    def draw_calibration_start(
            self, drawing_area, content_full_size, read_only=False):
        drawer = Drawer(self.core, drawing_area, read_only)
        drawer.set_content_full_size(content_full_size)
        drawer.start()
        self.drawers[drawing_area] = drawer

    def draw_calibration_stop(self, drawing_area):
        if drawing_area not in self.drawers:
            return
        self.drawers.pop(drawing_area).stop()

    def draw_scan_start(self, drawing_area, scan_id=None):
        if drawing_area in self.drawers:
            drawer = self.drawers[drawing_area]
        elif scan_id is not None and scan_id in self.scan_id_to_drawers:
            drawer = self.scan_id_to_drawers[scan_id]
            self.drawers.pop(drawer.drawing_area, None)
            drawer.stop()
            drawer.drawing_area = drawing_area
            drawer.start()
        else:
            drawer = Drawer(self.core, drawing_area, read_only=True)

        drawer.scan_id = scan_id
        self.scan_id_to_drawers[scan_id] = drawer
        self.drawers[drawing_area] = drawer

    def draw_scan_stop(self, drawing_area):
        if drawing_area in self.drawers:
            self.drawers.pop(drawing_area).stop()

    def on_scan_page_start(self, scan_id, page_nb, scan_params):
        for drawer in self.drawers.values():
            if drawer.scan_id is None or scan_id == drawer.scan_id:
                drawer.set_content_full_size(
                    (scan_params.get_width(), scan_params.get_height())
                )
                drawer.start()



import logging

try:
    import gi
    gi.require_version('Gdk', '3.0')
    from gi.repository import Gdk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class PageHoverHandler(object):
    def __init__(self, core, page):
        self.core = core
        self.page = page
        self.boxes = None
        self.actives = []

        self.realize_handler_id = None
        self.motion_handler_id = None

    def set_boxes(self, boxes, spatial_index):
        self.boxes = spatial_index

    def connect(self):
        assert self.realize_handler_id is None
        assert self.motion_handler_id is None
        self.realize_handler_id = self.page.widget.connect(
            "realize", self.on_realize
        )
        self.motion_handler_id = self.page.widget.connect(
            "motion-notify-event", self.on_motion
        )
        self.on_realize()

    def disconnect(self):
        assert self.realize_handler_id is not None
        assert self.motion_handler_id is not None
        self.page.widget.disconnect(self.realize_handler_id)
        self.page.widget.disconnect(self.motion_handler_id)
        self.realize_handler_id = None
        self.motion_handler_id = None

    def on_realize(self, widget=None):
        mask = Gdk.EventMask.POINTER_MOTION_MASK
        self.page.widget.add_events(mask)
        if self.page.widget.get_window() is not None:
            self.page.widget.get_window().set_events(
                self.page.widget.get_window().get_events() | mask
            )

    def on_motion(self, widget, event):
        if self.boxes is None:
            self.actives = []
            return
        zoom = self.page.get_zoom()
        x = int(event.x / zoom)
        y = int(event.y / zoom)
        actives = [(a[0], a[1].box) for a in self.boxes.get_boxes(x, y)]

        if len(actives) > 1:
            # will sort smaller areas last
            actives = [
                (abs((p[1][0] - p[0][0]) * (p[1][1] - p[0][1])), b)
                for (p, b) in actives
            ]
            actives.sort(reverse=True)

        actives = [
            a[1] for a in actives
        ]

        if actives != self.actives:
            self.page.widget.queue_draw()
        self.actives = actives

    def draw(self, cairo_ctx):
        for box in self.actives:
            self.core.call_all(
                "page_draw_box",
                cairo_ctx, self.page,
                box.position, (0.0, 0.0, 1.0),
                border_width=2, box_content=box.content
            )


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10

    def __init__(self):
        super().__init__()
        self.handlers = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview_boxes_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_pageview',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageview'],
            },
            {
                'interface': 'gtk_pageview_boxes',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageview.boxes'
                ],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_docview_closed_page(self, page):
        self.handlers.pop(page.widget).disconnect()

    def on_page_boxes_loaded(self, page, boxes, spatial_index):
        h = PageHoverHandler(self.core, page)
        self.handlers[page.widget] = h

        assert page.widget in self.handlers
        h = self.handlers[page.widget]
        h.set_boxes(boxes, spatial_index)
        if page.get_visible():
            h.connect()

    def on_page_visibility_changed(self, page, visible):
        if page.widget not in self.handlers:
            return
        h = self.handlers[page.widget]
        if visible:
            h.connect()
        else:
            h.disconnect()

    def on_page_draw(self, cairo_ctx, page):
        if page.widget not in self.handlers:
            return
        self.handlers[page.widget].draw(cairo_ctx)



import openpaperwork_core

from . import _version


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['app']

    def app_get_name(self):
        return "Paperwork"

    def app_get_fs_name(self):
        return "paperwork2"

    def app_get_version(self):
        return _version.version



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_redo_ocr_many"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -40

    def get_interfaces(self):
        return [
            'action',
            'action_docs',
            'action_docs_redo_ocr',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_selection',
                'defaults': ['paperwork_gtk.doc_selection'],
            },
            {
                'interface': 'ocr',
                'defaults': ['paperwork_backend.guesswork.ocr.pyocr'],
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._redo_ocr_many)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def _redo_ocr_single(
            self, doc_id, doc_url,
            progression_start_page_idx, progression_total_pages):
        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            LOGGER.warning("No pages in document %s. Nothing to do", doc_id)
            return

        LOGGER.info(
            "Will redo OCR on document %s (nb pages = %d)", doc_id, nb_pages
        )

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all,
            args=(
                "on_progress", "redo_ocr",
                progression_start_page_idx / progression_total_pages,
                _("OCR on %s") % doc_id)
        )
        promise = promise.then(lambda *args, **kwargs: None)

        for page_idx in range(0, nb_pages):
            promise = promise.then(
                self.core.call_all, "on_progress", "redo_ocr",
                (progression_start_page_idx + page_idx) /
                progression_total_pages,
                _("OCR on {doc_id} p{page_idx}").format(
                    doc_id=doc_id, page_idx=(page_idx + 1)
                )
            )
            promise = promise.then(lambda *args, **kwargs: None)
            promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
                self.core, self.core.call_all,
                args=("ocr_page_by_url", doc_url, page_idx,)
            ))
            promise = promise.then(lambda *args, **kwargs: None)

        promise = promise.then(
            self.core.call_all, "on_progress", "redo_ocr",
            (progression_start_page_idx + nb_pages) /
            progression_total_pages,
        )
        promise = promise.then(lambda *args, **kwargs: None)

        promise = promise.then(self.core.call_success(
            "transaction_simple_promise", [('upd', doc_id),]
        ))
        return (promise, nb_pages)

    def _redo_ocr_many(self, action, parameter):
        docs = set()
        self.core.call_all("doc_selection_get", docs)

        total_pages = 0
        for (doc_id, doc_url) in docs:
            nb_pages = self.core.call_success(
                "doc_get_nb_pages_by_url", doc_url
            )
            if nb_pages is None:
                nb_pages = 0
            total_pages += nb_pages

        promise = openpaperwork_core.promise.Promise(self.core)
        nb_all_pages = 0
        for (doc_id, doc_url) in docs:
            (p, nb_doc_pages) = self._redo_ocr_single(
                doc_id, doc_url, nb_all_pages, total_pages
            )
            nb_all_pages += nb_doc_pages
            promise = promise.then(p)

        self.core.call_success("transaction_schedule", promise)



import logging

import openpaperwork_core
import paperwork_backend.sync


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.page_info = None
        self.active_doc = (None, None)
        self.nb_pages = None
        self.current_page = None
        self.nb_pages = None

    def get_interfaces(self):
        return [
            'doc_open',
            'gtk_docview_pageinfo',
            'syncable',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow.docview.pageinfo", "pageinfo.css"
        )

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageinfo", "pageinfo.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.page_info = self.widget_tree.get_object("page_info")
        self.current_page = self.widget_tree.get_object(
            "page_current_nb"
        )
        self.current_page.connect("activate", self._change_page)
        self.nb_pages = self.widget_tree.get_object("page_total")

        button = self.widget_tree.get_object("page_prev")
        button.connect(
            "clicked",
            lambda *args, **kwargs: self.core.call_all(
                "doc_goto_previous_page"
            )
        )
        button = self.widget_tree.get_object("page_next")
        button.connect(
            "clicked",
            lambda *args, **kwargs: self.core.call_all("doc_goto_next_page")
        )

        self.core.call_success("docview_get_body").add_overlay(self.page_info)

        def _set_docview_margin(page_info, size_allocation):
            self.core.call_all(
                "docview_set_bottom_margin", page_info.get_allocated_height()
            )

        self.page_info.connect("size-allocate", _set_docview_margin)

    def _change_page(self, *args, **kwargs):
        txt = self.current_page.get_text()
        if txt == "":
            return
        self.core.call_all("doc_goto_page", int(txt) - 1)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            nb_pages = 0

        self.current_page.set_text("")
        self.nb_pages.set_text(f"/ {nb_pages}")
        self.page_info.set_visible(True)
        self.page_info.set_sensitive(nb_pages > 0)

    def doc_reload(self, doc_id, doc_url):
        if (doc_id, doc_url) != self.active_doc:
            return
        self.doc_open(*self.active_doc)

    def on_page_shown(self, page_idx):
        txt = str(page_idx + 1)
        if txt != self.current_page.get_text():
            self.current_page.set_text(txt)

    def page_info_add_left(self, widget):
        self.widget_tree.get_object("page_prevnext").pack_end(
            widget, expand=False, fill=True, padding=0
        )
        return True

    def page_info_add_right(self, widget):
        self.widget_tree.get_object("page_info").pack_end(
            widget, expand=False, fill=True, padding=0
        )
        return True

    def doc_transaction_start(self, out: list, total_expected=-1):
        class RefreshTransaction(paperwork_backend.sync.BaseTransaction):
            priority = -100000

            def __init__(s, core, total_expected=-1):
                super().__init__(core, total_expected)
                s.active_doc = False

            def _refresh(s):
                self.core.call_success(
                    "mainloop_schedule", self.doc_open, *self.active_doc
                )
                s.active_doc = True

            def add_doc(s, doc_id):
                if self.active_doc[0] == doc_id:
                    s._refresh()

            def upd_doc(s, doc_id):
                if self.active_doc[0] == doc_id:
                    s._refresh()

            def del_doc(s, doc_id):
                if self.active_doc[0] == doc_id:
                    s._refresh()

            def cancel(s):
                if s.active_doc:
                    s._refresh()

            def commit(s):
                if s.active_doc:
                    s._refresh()

        out.append(RefreshTransaction(self.core, total_expected))

    def sync(self, promises: list):
        # sync don't change document content --> no need to refresh
        pass



"""
Let other plugins replace a page image without actually smashing the original
image. Also provide a method to drop the modified version of a page and revert
to the original only.
"""

import logging
import re

import openpaperwork_core

from . import util


LOGGER = logging.getLogger(__name__)

PAGE_FILENAME_PNG_FMT = "paper.{}.edited.png"
PAGE_FILENAME_JPG_FMT = "paper.{}.edited.jpg"
PAGE_FILENAME_FMTS = [
    PAGE_FILENAME_PNG_FMT,
    PAGE_FILENAME_JPG_FMT,
]
PAGE_FILENAME_REGEX = re.compile(r"paper\.(\d+)\.edited.(jpg|png)")


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000  # must have a higher priority than model.img / model.pdf

    def get_interfaces(self):
        return [
            "page_img",
            "page_reset",
            'pages',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
        ]

    def init(self, core):
        self.core = core
        setting = self.core.call_success(
            "config_build_simple", "model",
            "img_overlay_format", lambda: "PNG"
        )
        self.core.call_all(
            "config_register", "model_img_overlay_format", setting
        )

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        mtime = util.get_doc_mtime(self.core, doc_url, PAGE_FILENAME_REGEX)
        if mtime is not None:
            out.append(mtime)

    def page_internal_get_mtime_by_url(self, out: list, doc_url, page_idx):
        for filename_fmt in PAGE_FILENAME_FMTS:
            mtime = util.get_page_mtime(
                self.core, doc_url, page_idx, filename_fmt
            )
            if mtime is not None:
                out.append(mtime)
                break

    def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
        for filename_fmt in PAGE_FILENAME_FMTS:
            h = util.get_page_hash(
                self.core, doc_url, page_idx, filename_fmt)
            if h is not None:
                out.append(h)
                break

    def page_get_img_url(
                self, doc_url, page_idx, write=False, original=False,
                **kwargs
            ):
        if original:
            # caller want the URL of the original image, not the edited one
            return None

        for filename_fmt in PAGE_FILENAME_FMTS:
            page_url = self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(page_idx + 1)
            )
            if self.core.call_success("fs_exists", page_url) is not None:
                return page_url

        if not write:
            return None

        # caller wants to modify or create a page
        # check if we already have an original image. If so, we return our
        # URL. Otherwise, we let the caller write the original image first.
        if self.core.call_success(
                    "page_get_img_url", doc_url, page_idx, write=False
                ) is not None:
            # has already an original page (or even an edited one)
            # --> return our URL
            img_fmt = self.core.call_success(
                "config_get", "model_img_overlay_format"
            )
            if img_fmt == "PNG":
                filename_fmt = PAGE_FILENAME_PNG_FMT
            else:
                filename_fmt = PAGE_FILENAME_JPG_FMT
            return self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(page_idx + 1)
            )

        # let the caller write an original page (see model.img)
        return None

    def page_delete_by_url(self, doc_url, page_idx):
        r = None
        for filename_fmt in PAGE_FILENAME_FMTS:
            r = util.delete_page_file(
                self.core, filename_fmt, doc_url, page_idx
            ) or r
        return r

    def page_move_by_url(
                self,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ):
        r = None
        for filename_fmt in PAGE_FILENAME_FMTS:
            r = util.move_page_file(
                self.core, filename_fmt,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ) or r
        return r

    def page_reset_by_url(self, doc_url, page_idx):
        """
        Reset a page image to its original content. In other word, we simply
        delete the edited image so the original one takes over again
        (see model.img).
        """
        for filename_fmt in PAGE_FILENAME_FMTS:
            page_url = self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(page_idx + 1)
            )
            if self.core.call_success("fs_exists", page_url):
                self.core.call_success("fs_unlink", page_url, trash=False)



import asyncio
import logging
import sys
import threading

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    """
    A main loop based on asyncio. Not as complete as GLib main loop, but
    good enough for shell commands.
    """
    def __init__(self):
        super().__init__()
        self.halt_on_uncaught_exception = True
        self.loop = None
        self.loop_ident = None
        self.halt_cause = None
        self.task_count = 0
        self.log_uncaught = True

    def _check_mainloop_instantiated(self):
        if self.loop is None:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            self.task_count = 0

    def get_interfaces(self):
        return [
            "mainloop",
        ]

    def mainloop(self, halt_on_uncaught_exception=True, log_uncaught=True):
        """
        Wait for callbacks to be scheduled and execute them.

        This method is blocking and will block until `mainloop_quit*()` is
        called.
        """
        self._check_mainloop_instantiated()
        self.log_uncaught = log_uncaught
        self.halt_on_uncaught_exception = halt_on_uncaught_exception

        self.mainloop_schedule(self.core.call_all, "on_mainloop_start")

        self.loop_ident = threading.current_thread().ident
        try:
            self.loop.run_forever()
        finally:
            self.loop_ident = None

        self.core.call_all("on_mainloop_quit")

        if self.halt_cause is not None:
            halt_cause = self.halt_cause
            self.halt_cause = None
            LOGGER.error("Main loop stopped because %s", str(halt_cause))
            raise halt_cause

        self.loop = None
        return True

    def mainloop_get_thread_id(self):
        """
        Gets the ID of the thread running the main loop. `None` if no
        thread is running it.
        """
        return self.loop_ident

    def mainloop_quit_graceful(self):
        """
        Wait for all the scheduled callbacks to be executed and then stops
        the main loop.
        """
        self.mainloop_schedule(self._mainloop_quit_graceful)
        return True

    def _mainloop_quit_graceful(self):
        if self.task_count > 1:  # keep in mind this function is in a task too
            LOGGER.info(
                "Quit graceful: Remaining tasks: %d", self.task_count - 1
            )
            self.mainloop_schedule(self._mainloop_quit_graceful, delay_s=0.2)
            return

        LOGGER.info("Quit graceful: Quitting")
        self.mainloop_quit_now()
        self.task_count = 1  # we are actually the one task still running

    def mainloop_quit_now(self):
        """
        Stops the main loop right now.

        Note that it cannot interrupt a callback being executed, but no
        callback scheduled after this one will be executed.
        """
        if self.loop is None:
            return None
        self.loop.stop()
        self.loop = None
        self.task_count = 0

    def mainloop_ref(self, obj):
        """
        If you run a task independently from the main loop, you may want
        to increment the reference counter of the main loop so
        `mainloop_quit_graceful` does not interrupt the main loop while your
        task is still running.

        ThreadedPromise already takes care of incrementing and decrementing
        this reference counter.
        """
        self.task_count += 1

    def mainloop_unref(self, obj):
        self.task_count -= 1

    def mainloop_schedule(self, func, *args, delay_s=0, **kwargs):
        """
        Request that the main loop executes the callback `func`.
        Will return immediately.
        """
        assert hasattr(func, '__call__')

        self._check_mainloop_instantiated()

        self.task_count += 1

        async def decorator(args, kwargs):
            try:
                func(*args, **kwargs)
            except Exception as exc:
                exc_info = sys.exc_info()
                if self.halt_on_uncaught_exception:
                    LOGGER.error(
                        "Main loop: Uncaught exception (%s) ! Quitting",
                        func, exc_info=exc
                    )
                    self.halt_cause = exc
                    self.mainloop_quit_now()
                elif self.log_uncaught:
                    LOGGER.error(
                        "Main loop: Uncaught exception (%s) !",
                        func, exc_info=exc
                    )
                self.core.call_all(
                    "mainloop_schedule",
                    self.core.call_all, "on_uncaught_exception", exc_info
                )
            finally:
                self.task_count -= 1

        coroutine = decorator(args, kwargs)

        args = (self.loop.create_task, coroutine)
        if delay_s != 0:
            args = (self.loop.call_later, delay_s) + args
        self.loop.call_soon_threadsafe(args[0], *(args[1:]))
        return True

    def mainloop_execute(self, func, *args, **kwargs):
        """
        Ensure a function is run on the main loop, even if called from
        a thread. Will return only once the callback `func` has been
        executed. Will return the value returned by `func`.

        This method makes it easier to work with non-thread-safe modules
        (sqlite3 for instance).
        """
        current = threading.current_thread().ident

        # XXX(Jflesch):
        # if self.loop_ident is None, it means the mainloop hasn't been started
        # yet --> we cannot run the function on the mainloop anyway, so
        # we assume we are on the same thread that will later run the main
        # loop.
        if self.loop_ident is None or current == self.loop_ident:
            return func(*args, **kwargs)

        event = threading.Event()
        out = None
        exc = None

        def get_result():
            nonlocal out
            nonlocal exc
            try:
                out = func(*args, **kwargs)
            except Exception as e:
                exc = e
            event.set()

        self.mainloop_schedule(get_result)
        event.wait()

        if exc is not None:
            raise exc
        return out



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_doclist_listener',
            'gtk_doclist_name',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.maindow.doclist'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_doc_box_creation(self, doc_id, gtk_row, custom_flowlayout):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        is_new = doc_url is None or self.core.call_success(
            "is_doc", doc_url
        ) is None
        if is_new:
            doc_txt = _("New document")
        else:
            doc_date = self.core.call_success("doc_get_date_by_id", doc_id)
            if doc_date is not None:
                doc_txt = self.core.call_success("i18n_date_short", doc_date)
            else:
                doc_txt = doc_id
        label = Gtk.Label.new(doc_txt)
        label.set_visible(True)
        custom_flowlayout.add_child(label, Gtk.Align.START)



import logging

import openpaperwork_core
import openpaperwork_core.promise

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    RECOMMENDED = 300

    def get_interfaces(self):
        return [
            'gtk_settings_scanner_setting',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_settings_scanner',
                'defaults': ['paperwork_gtk.settings.scanner.settings'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def complete_scanner_settings(
            self, global_widget_tree, parent_widget_tree,
            list_scanner_promise):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings.scanner",
            "popover.glade"
        )

        selector = widget_tree.get_object("selector")

        # WORKAROUND(Jflesch): set_sensitive() doesn't appear to work on
        # GtkMenuButton --> we have to play with set_popover()

        def reset_popover():
            dev_id = self.core.call_success("config_get", "scanner_dev_id")
            parent_widget_tree.get_object("scanner_resolution").set_popover(
                selector if dev_id is not None and dev_id != "" else None
            )

        reset_popover()
        self.core.call_all(
            "config_add_observer", "scanner_dev_id", reset_popover
        )

        selector.connect("show", self._on_show, widget_tree)

    def _on_show(self, popover, widget_tree):
        LOGGER.info("Scanner resolution selector is visible")

        widget_tree.get_object("settings_stack").set_visible_child_name(
            "spinner"
        )
        widget_tree.get_object("spinner").start()
        box = widget_tree.get_object("selector_box")
        for child in box.get_children():
            box.remove(child)

        dev_id = self.core.call_success("config_get", "scanner_dev_id")
        if dev_id is None:
            # TODO(Jflesch): better display
            self._display_resolutions([], widget_tree)
            return
        promise = self.core.call_success("scan_get_scanner_promise", dev_id)
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, self._collect_resolutions
        ))
        promise = promise.then(self._display_resolutions, widget_tree)
        promise = promise.catch(self._on_error, widget_tree)
        self.core.call_success("scan_schedule", promise)

    def _collect_resolutions(self, dev):
        resolutions = set()
        try:
            children = dev.dev.get_children()
            for child in children:
                opts = {o.get_name(): o for o in child.get_options()}
                if 'resolution' not in opts:
                    continue
                constraint = opts['resolution'].get_constraint()
                resolutions.update(constraint)
        finally:
            dev.close()
        LOGGER.info("Got resolutions: %s", resolutions)
        resolutions = list(resolutions)
        resolutions.sort()
        return resolutions

    def _display_resolutions(self, resolutions, widget_tree):
        widget_tree.get_object("spinner").stop()
        widget_tree.get_object("settings_stack").set_visible_child_name(
            "selector"
        )

        box = widget_tree.get_object("selector_box")
        radios = []
        for resolution in resolutions:
            radio = self.core.call_success(
                "gtk_load_widget_tree", "paperwork_gtk.settings.scanner",
                "popover_box.glade"
            )
            radio = radio.get_object("radio")
            if resolution != self.RECOMMENDED:
                radio.set_label(_("{} dpi").format(resolution))
            else:
                radio.set_label(_("{} dpi (recommended)").format(resolution))
            box.pack_start(radio, expand=False, fill=True, padding=0)
            radios.append((resolution, radio))

        for (resolution, radio) in radios[1:]:
            radio.join_group(radios[0][1])

        active = self.core.call_success("config_get", "scanner_resolution")
        for (resolution, radio) in radios:
            if active == resolution:
                radio.set_active(True)

        for (resolution, radio) in radios:
            radio.connect(
                "toggled", self._on_toggle, widget_tree, resolution
            )

    def _on_toggle(self, checkbox, widget_tree, resolution):
        LOGGER.info("Selected resolution: %d", resolution)
        widget_tree.get_object("selector").popdown()
        self.core.call_success("config_put", "scanner_resolution", resolution)

    def _on_error(self, exc, widget_tree):
        LOGGER.error("Fail to get scanner resolutions", exc_info=exc)
        # TODO(Jflesch): better display
        widget_tree.get_object("spinner").stop()
        widget_tree.get_object("settings_stack").set_visible_child_name(
            "selector"
        )



import logging

import openpaperwork_core
import openpaperwork_core.deps

from ... import _

LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -500

    def get_interfaces(self):
        return [
            'gtk_settings',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'i18n_lang',
                'defaults': ['paperwork_backend.i18n.pycountry'],
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def complete_settings(self, global_widget_tree):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings.ocr",
            "settings.glade"
        )

        label = widget_tree.get_object("ocr_langs_label")
        self._update_langs(label)

        self.core.call_all("complete_ocr_settings", widget_tree)

        def refresh(*args, **kwargs):
            self._update_langs(label)

        def disable_refresh(*args, **kwargs):
            self.core.call_all("config_remove_observer", "ocr_langs", refresh)

        self.core.call_all("config_add_observer", "ocr_langs", refresh)

        global_widget_tree.get_object("settings_window").connect(
            "destroy", disable_refresh
        )

        self.core.call_success(
            "add_setting_to_dialog", global_widget_tree,
            _("Optical Character Recognition"),
            [widget_tree.get_object("ocr_langs")]
        )

    def _update_langs(self, label):
        langs = self.core.call_success("ocr_get_active_langs")
        langs = [
            self.core.call_success("i18n_lang_iso639_3_to_full", lang)
            for lang in langs
        ]
        langs = ", ".join(langs)
        if langs == "":
            langs = _("OCR disabled")
        LOGGER.info("OCR languages: %s", langs)
        label.set_text(langs)



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 750

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.active_doc = None

    def get_interfaces(self):
        return [
            'gtk_doc_property',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'extra_text',
                'defaults': ['paperwork_backend.model.extra_text'],
            },
            {
                'interface': 'gtk_doc_properties',
                'defaults': ['paperwork_gtk.docproperties'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def _get_widget_text(self):
        text_buffer = self.widget_tree.get_object("doctext_text").get_buffer()
        start = text_buffer.get_iter_at_offset(0)
        end = text_buffer.get_iter_at_offset(-1)
        return text_buffer.get_text(start, end, False)

    def doc_properties_components_get(self, out: list, multiple_docs=False):
        if multiple_docs:
            return
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docproperties", "extra_text.glade"
        )

        out.append(self.widget_tree.get_object("doctext"))

    def doc_properties_components_set_active_doc(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)
        txt = []
        self.core.call_all("doc_get_extra_text_by_url", txt, doc_url)
        txt = "\n".join(txt)
        self.widget_tree.get_object("doctext_text").get_buffer().set_text(txt)

    def doc_properties_components_apply_changes(self, out):
        if out.multiple_docs:
            return

        # The document may have been renamed: use out.doc_id instead of
        # self.active_doc
        doc_id = out.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        self.active_doc = (doc_id, doc_url)

        orig_txt = []
        self.core.call_all(
            "doc_get_extra_text_by_url", orig_txt, doc_url
        )
        orig_txt = "\n".join(orig_txt).strip()

        new_txt = self._get_widget_text()

        if new_txt == orig_txt:
            return

        LOGGER.info("Extra keywords have been changed in document %s", doc_id)
        self.core.call_all(
            "doc_set_extra_text_by_url", doc_url, new_txt
        )

        out.upd_docs.add(doc_id)

    def doc_properties_components_cancel_changes(self):
        self.doc_properties_components_set_active_doc(*self.active_doc)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("doctext_text"),
            self.core.call_success(
                "fs_join", out_dir, "doc_extra_text.png"
            ),
            margins=(50, 50, 50, 50)
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import rich.text

import openpaperwork_core

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "document_storage",
                "defaults": ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'rename', help=_("Change a document identifier")
        )
        p.add_argument(
            'source_doc_id',
            help=_("Document to rename")
        )
        p.add_argument(
            'dest_doc_id',
            help=_("New name for the document")
        )

    def cmd_run(self, console, args):
        if args.command != 'rename':
            return None

        source_doc_id = args.source_doc_id
        dest_doc_id = args.dest_doc_id

        source_doc_url = self.core.call_success("doc_id_to_url", source_doc_id)
        dest_doc_url = self.core.call_success(
            "doc_id_to_url", dest_doc_id, existing=False
        )

        console.print(rich.text.Text(
            f"Renaming: {source_doc_url} --> {dest_doc_url}"
        ))

        self.core.call_all("doc_rename_by_url", source_doc_url, dest_doc_url)
        self.core.call_success(
            "transaction_simple",
            [
                ('del', source_doc_id),
                ('add', dest_doc_id),
            ]
        )
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

        console.print(rich.text.Text(
            f"{source_doc_id} renamed into {dest_doc_id}"
        ))

        return True



from openpaperwork_core import (
    _,
    PluginBase
)


class Plugin(PluginBase):
    def __init__(self):
        self.console = None

    def get_interfaces(self):
        return ['shell']

    def cmd_complete_argparse(self, parser):
        parser.add_parser(
            "ping",
            help=_("Check that all required dependencies are installed")
        )

    def cmd_run(self, console, args):
        if args.command != 'ping':
            return None
        console.print("Pong")
        return True



import logging

import openpaperwork_core
import openpaperwork_core.promise
import paperwork_backend.sync


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1000000

    def __init__(self):
        super().__init__()
        self.search_entry = None
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'gtk_search_field',
            'screenshot_provider',
            'syncable',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'index',
                'defaults': ['paperwork_backend.index.whoosh'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.search", "field.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.search_entry = self.widget_tree.get_object("search_entry")
        search_field = self.widget_tree.get_object("search_field")

        self.core.call_all("doclist_add", search_field, vposition=0)

        self.search_entry.connect(
            "search-changed", self.search_update_document_list
        )
        self.search_entry.connect("stop-search", lambda w: self.search_stop())

        self.widget_tree.get_object("search_dialog_button").connect(
            "clicked", lambda button: self.core.call_all(
                "gtk_open_advanced_search_dialog"
            )
        )

        self.core.call_all("work_queue_create", "doc_search")
        self.core.call_one(
            "mainloop_schedule", self.search_update_document_list
        )

    def search_update_document_list(self, _=None):
        query = self.search_entry.get_text()
        LOGGER.info("Looking for [%s]", query)
        self.core.call_all("search_by_keywords", query)

    def search_get(self):
        return self.search_entry.get_text()

    def search_set(self, text):
        self.search_entry.set_text(text)

    def search_by_keywords(self, query):
        self.core.call_all("work_queue_cancel_all", "doc_search")
        self.core.call_all("on_search_start", query)
        if query == "":
            out = []
            promise = openpaperwork_core.promise.ThreadedPromise(
                self.core, lambda: self.core.call_all(
                    "storage_get_all_docs", out
                )
            )
        else:
            out = []
            promise = openpaperwork_core.promise.ThreadedPromise(
                self.core, lambda: self.core.call_all(
                    "index_search", out, query
                )
            )
        promise = promise.then(lambda *args, **kwargs: out)
        promise = promise.then(lambda docs: sorted(docs, reverse=True))

        def show_if_query_still_valid(docs):
            # While we were looking for the documents, the query may have
            # changed (user tying). No point in displaying obsolete results.
            if query != self.search_entry.get_text():
                return
            self.core.call_all("on_search_results", query, docs)

        promise = promise.then(show_if_query_still_valid)

        self.core.call_all("work_queue_add_promise", "doc_search", promise)

    def search_stop(self):
        self.core.call_all("work_queue_cancel_all", "doc_search")

    def doc_transaction_start(self, out: list, total_expected=-1):
        class RefreshTransaction(paperwork_backend.sync.BaseTransaction):
            priority = -100000

            def commit(s):
                self.core.call_one(
                    "mainloop_schedule", self.search_update_document_list
                )

        out.append(RefreshTransaction(self.core, total_expected))

    def sync(self, promises: list):
        # If someone requested a sync, assume something has changed
        # --> try to keep the view as up-to-date as possible.
        self.search_update_document_list()

        promises.append(openpaperwork_core.promise.Promise(
            self.core, self.search_update_document_list
        ))

    def screenshot_snap_all_doc_widgets(self, out_dir):
        self.core.call_success(
            "screenshot_snap_widget", self.search_entry,
            self.core.call_success("fs_join", out_dir, "search.png"),
            margins=(50, 50, 50, 140)
        )
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("search_dialog_button"),
            self.core.call_success(
                "fs_join", out_dir, "advanced_search_button.png"
            ),
            margins=(200, 30, 30, 30)
        )

    def search_focus(self):
        LOGGER.info("Focusing on search field")
        self.widget_tree.get_object("search_entry").grab_focus()

    def search_field_add(self, widget):
        search_field = self.widget_tree.get_object("search_field")
        search_field.add(widget)



import logging

import openpaperwork_core

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    """
    Label files in each document contains both the label names and their color.
    If the user changes a label color, label files are updated one-by-one.
    If this process is interrupted, we end up with 2 colors for the same
    label.
    """
    def get_interfaces(self):
        return ['chkworkdir']

    def get_deps(self):
        return [
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
        ]

    @staticmethod
    def _is_same_color(a, b):
        a = (int(a[0] * 0xFF), int(a[1] * 0xFF), int(a[2] * 0xFF))
        b = (int(b[0] * 0xFF), int(b[1] * 0xFF), int(b[2] * 0xFF))
        return a == b

    def check_work_dir(self, out_problems: list):
        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs, only_valid=False)
        all_docs.sort()

        total = len(all_docs)
        LOGGER.info("Checking work directory (%d documents)", total)

        # label text --> (
        #     label color, first_seen_doc_id, first_seen_exact_color
        # )
        first_seen_labels = {}

        for (idx, (doc_id, doc_url)) in enumerate(all_docs):
            self.core.call_all(
                "on_progress", "chkworkdir_label_color",
                idx / total, _("Checking doc %s") % (doc_id,)
            )

            doc_labels = set()
            self.core.call_all("doc_get_labels_by_url", doc_labels, doc_url)
            for (doc_label_txt, doc_label_color_orig) in doc_labels:
                doc_label_color = self.core.call_success(
                    "label_color_to_rgb", doc_label_color_orig
                )

                first_color = first_seen_labels.get(doc_label_txt, None)
                if first_color is None:
                    first_seen_labels[doc_label_txt] = (
                        doc_label_color, doc_id, doc_label_color_orig
                    )
                    continue

                (first_color, first_docid, first_exact_color) = first_color
                if self._is_same_color(first_color, doc_label_color):
                    continue

                out_problems.append({
                    "problem": "label_color",
                    "doc_id": doc_id,
                    "doc_url": doc_url,
                    "problem_color": doc_label_color,
                    "solution_color": first_color,
                    "current_label": (doc_label_txt, doc_label_color),
                    "fixed_label": (doc_label_txt, first_exact_color),
                    "human_description": {
                        "problem": (
                            _(
                                "Document %s has label \"%s\" with color=%s"
                                " while document %s has label"
                                " \"%s\" with color=%s"
                            ) % (
                                doc_id, doc_label_txt,
                                self.core.call_success(
                                    "label_color_rgb_to_text",
                                    doc_label_color
                                ),
                                first_docid,
                                doc_label_txt,
                                self.core.call_success(
                                    "label_color_rgb_to_text",
                                    first_color
                                )
                            )
                        ),
                        "solution": (
                            _(
                                "Set label color %s on label \"%s\""
                                " of document %s"
                            )
                            % (
                                self.core.call_success(
                                    "label_color_rgb_to_text",
                                    first_color
                                ),
                                doc_label_txt, doc_id
                            )
                        )
                    }
                })

        self.core.call_all("on_progress", "chkworkdir_label_color", 1.0)

    def fix_work_dir(self, problems):
        total = len(problems)
        for (idx, problem) in enumerate(problems):
            if problem['problem'] != 'label_color':
                continue
            LOGGER.info("Fixing document %s", problem['doc_url'])
            self.core.call_all(
                "on_progress", "fixworkdir_label_color",
                idx / total, _("Fixing label on doc %s") % (problem['doc_id'],)
            )
            self.core.call_success(
                "doc_remove_label_by_url", problem['doc_url'],
                problem['current_label'][0]
            )
            self.core.call_success(
                "doc_add_label_by_url", problem['doc_url'],
                problem['fixed_label'][0], problem['fixed_label'][1]
            )
        self.core.call_all("on_progress", "fixworkdir_label_color", 1.0)



try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core

from . import deps


class Plugin(openpaperwork_core.PluginBase):
    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_colors',
        ]

    def gtk_entry_set_colors(self, gtk_entry, fg="black", bg="#CC3030"):
        css = """
        * {
            color: %s;
            background: %s;
        }

        * selection {
            color: white;
            background: #3b84e9;
        }
        """ % (fg, bg)
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(css.encode())
        css_context = gtk_entry.get_style_context()
        css_context.add_provider(
            css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER
        )

    def gtk_entry_reset_colors(self, gtk_entry):
        self.gtk_entry_set_colors(
            gtk_entry, fg="@theme_text_color", bg="@theme_bg_color"
        )

    def gtk_theme_get_color(self, color_name):
        style = Gtk.StyleContext()
        return style.lookup_color("theme_bg_color")[1]



import logging

import openpaperwork_core

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    """
    Locates (and optionnally deletes) empty directories in the work
    directory.
    """
    def get_interfaces(self):
        return ['chkworkdir']

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'nb_pages',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
        ]

    def check_work_dir(self, out_problems: list):
        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs, only_valid=False)
        all_docs.sort()

        total = len(all_docs)
        LOGGER.info("Checking work directory (%d documents)", total)
        for (idx, (doc_id, doc_url)) in enumerate(all_docs):
            self.core.call_all(
                "on_progress", "chkworkdir_empty_doc",
                idx / total, _("Checking doc %s") % (doc_id,)
            )
            isdir = self.core.call_success("fs_isdir", doc_url)
            if not isdir:
                continue
            try:
                nb_pages = self.core.call_success(
                    "doc_get_nb_pages_by_url", doc_url
                )
            except Exception:
                # handled by 'paperwork_backend.chkworkdir.corrupted_page_map'
                continue
            if nb_pages is not None and nb_pages > 0:
                continue
            out_problems.append({
                "problem": "empty_doc",
                "doc_id": doc_id,
                "doc_url": doc_url,
                "human_description": {
                    "problem": _("Document %s is empty") % (doc_id,),
                    "solution": _("Delete document %s") % (doc_id,),
                },
            })

        self.core.call_all("on_progress", "chkworkdir_empty_doc", 1.0)

    def fix_work_dir(self, problems):
        total = len(problems)
        for (idx, problem) in enumerate(problems):
            if problem['problem'] != 'empty_doc':
                continue
            LOGGER.info("Fixing document %s", problem['doc_url'])
            self.core.call_all(
                "on_progress", "fixworkdir_empty_doc",
                idx / total, _("Deleting empty doc %s") % (problem['doc_id'],)
            )
            self.core.call_all("storage_delete_doc_id", problem['doc_id'])
        self.core.call_all("on_progress", "fixworkdir_empty_doc", 1.0)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'action',
            'action_doc',
            'action_doc_new',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_open',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        action = Gio.SimpleAction.new("doc_new", None)
        action.connect("activate", self._open_new_doc)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def _open_new_doc(self, *args, **kwargs):
        self.core.call_all("doc_open_new")



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import openpaperwork_core

from . import util
from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "ocr",
                "defaults": ['paperwork_backend.guesswork.ocr.pyocr'],
            },
            {
                "interface": "ocr_settings",
                "defaults": ['paperwork_backend.pyocr'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'ocr', help=_(
                "OCR document or pages"
            )
        )
        p.add_argument(
            'doc_id', type=str,
            help=_("Document on which OCR must be run")
        )
        p.add_argument(
            '--pages', '-p', type=str,
            help=_(
                "Pages to OCR"
                " (single integer, range or comma-separated list,"
                " default: all pages)"
            )
        )

    def cmd_run(self, console, args):
        if args.command != 'ocr':
            return None

        if self.core.call_success("ocr_is_enabled") is None:
            console.print("OCR is disabled")
            return []

        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        pages = util.parse_page_list(args)
        if pages is None:
            nb_pages = self.core.call_success(
                "doc_get_nb_pages_by_url", doc_url
            )
            pages = range(0, nb_pages)

        out = []

        for page_idx in pages:
            console.print(
                _(
                    "Running OCR on"
                    " document {doc_id} page {page_idx} ..."
                ).format(
                    doc_id=doc_id, page_idx=(page_idx + 1)
                )
            )

            self.core.call_all("ocr_page_by_url", doc_url, page_idx)

            console.print(_("Done"))

            out.append((doc_id, page_idx))

        console.print(_("All done !"))

        return out



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging
import shutil
import sys

import rich.text

import openpaperwork_core

import paperwork_backend.pageedit

from . import util
from .. import _


LOGGER = logging.getLogger(__name__)


class NullUI(paperwork_backend.pageedit.AbstractPageEditorUI):
    def __init__(self):
        super().__init__()


class CliUI(paperwork_backend.pageedit.AbstractPageEditorUI):
    def __init__(self, core, console):
        super().__init__()
        self.core = core
        self.console = console

    def show_preview(self, img):
        terminal_width = shutil.get_terminal_size()[0] - 1
        img = self.core.call_success(
            "img_render", img, terminal_width=terminal_width
        )
        if img is None:
            return
        for line in img:
            self.console.print(rich.text.Text.from_ansi(line))
        self.console.print("")


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            # optional dependency
            # {
            #     "interface": "img_renderer",
            #     "defaults": ["paperwork_shell.display.docrendering.img"],
            # },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                "interface": "page_editor",
                "defaults": ["paperwork_backend.pageedit.pageeditor"],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        # just so we can get the modifier list
        editor = self.core.call_success("page_editor_get", None, 0, NullUI())
        modifiers = editor.get_modifiers()

        modifiers = [
            modifier['id']
            for modifier in modifiers
            if not modifier['need_frame']
        ]

        edit_parser = parser.add_parser('edit', help=_("Edit page"))
        edit_parser.add_argument(
            '--modifiers', '-m', type=str, required=True,
            help=_(
                "List of image modifiers (comma separated, possible values:"
                " {})"
            ).format(modifiers)
        )
        # we need page number(s), but for consistency with other commands,
        # we require this argument as an option like '--opt' instead of
        # a positional argument.
        edit_parser.add_argument('--pages', '-p', type=str, required=True)
        edit_parser.add_argument('doc_id')

    def cmd_run(self, console, args):
        if args.command != 'edit':
            return None
        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        pages = util.parse_page_list(args)
        modifiers = args.modifiers.split(",")
        out = []

        for page_idx in pages:
            out.append((doc_id, page_idx))
            console.print(
                _("Modifying document {} page {} ...").format(
                    doc_id, page_idx
                )
            )
            console.print(_("Original:"))
            ui = CliUI(self.core, console)

            page_editor = self.core.call_success(
                "page_editor_get", doc_url, page_idx, ui
            )

            promise = openpaperwork_core.promise.Promise(self.core)
            for modifier in modifiers:
                promise = promise.then(console.print, f"{modifier}:")
                promise = promise.then(
                    page_editor.on_modifier_selected(modifier)
                )

            if promise is not None:
                promise = promise.then(
                    sys.stdout.write,
                    _("Generating in high quality and saving ...") + " "
                )
                promise = promise.then(
                    lambda *args, **kwargs: sys.stdout.flush()
                )
                promise = promise.then(page_editor.on_save())
                promise = promise.then(console.print, "Done")

                def on_err(exc):
                    LOGGER.error(
                        "Edition of %s p%d failed",
                        doc_id, page_idx,
                        exc_info=exc
                    )
                    page_editor.on_cancel()
                    raise exc

                promise.catch(on_err)

                promise.schedule()

                self.core.call_all("mainloop_quit_graceful")
                self.core.call_one("mainloop")

        console.print(_("Committing ...") + " ")

        self.core.call_success("transaction_simple", [("upd", doc_id),])
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

        console.print(_("Done"))
        console.print(_("All done !"))

        return out



import datetime
import json
import logging
import uuid

import openpaperwork_core
import openpaperwork_core.promise

from . import PeriodicTask
from .. import _


LOGGER = logging.getLogger(__name__)

POST_STATS_INTERVAL = datetime.timedelta(days=7)
POST_STATS_PATH = "/beacon/post_statistics"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.periodic = None
        self.http = None

    def get_interfaces(self):
        return [
            "bug_report_attachments",
            "stats_post",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['paperwork_backend.app'],
            },
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'fs',
                'defaults': [
                    'openpaperwork_core.fs.memory',
                    'openpaperwork_core.fs.python',
                ],
            },
            {
                'interface': 'http_json',
                'defaults': ['openpaperwork_core.http'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.periodic = PeriodicTask(
            "statistics",
            datetime.timedelta(days=7),
            self.stats_send
        )
        self.http = self.core.call_success(
            "http_json_get_client", "statistics"
        )

        self._register_config(core)
        self.periodic.register_config(core)

        if self.core.call_success("config_get", "send_statistics"):
            self.periodic.do(core)

    def _register_config(self, core):
        setting = self.core.call_success(
            "config_build_simple", "statistics",
            "enabled", lambda: False
        )
        self.core.call_all(
            "config_register", "send_statistics", setting
        )
        setting = self.core.call_success(
            "config_build_simple", "statistics",
            "uuid", lambda: uuid.getnode()
        )
        self.core.call_all("config_register", "uuid", setting)

    def _collect_stats(self, node_uuid):
        stats = {
            'uuid': node_uuid,
            'paperwork_version': self.core.call_success("app_get_version"),
            'nb_documents': 0,
            'os_name': '',
            'platform_architecture': '',
            'platform_processor': '',
            'platform_distribution': '',
            'cpu_count': 0,
        }
        self.core.call_all("stats_get", stats)
        return {'statistics': stats}

    def stats_send(self):
        node_uuid = self.core.call_success("config_get", "uuid")
        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._collect_stats, args=(node_uuid,)
        )
        promise = promise.then(self.http.get_request_promise(POST_STATS_PATH))

        def on_request_done(reply):
            LOGGER.info("Statistics posted. Reply: {}".format(reply))
            self.core.call_all('on_stats_sent')

        promise = promise.then(on_request_done)
        promise = promise.catch(self._on_stats_send_error)
        promise.schedule()

    def _on_stats_send_error(self, exc):
        LOGGER.warning("Failed to send stats", exc_info=exc)

    def bug_report_get_attachments(self, out: dict):
        out['stats'] = {
            'include_by_default': True,
            'date': None,
            'file_type': _("App. & system info."),
            'file_url': _("Select to generate"),
            'file_size': 0,
        }

    def _write_stats_to_tmp_file(self, stats):
        stats = json.dumps(
            stats, indent=4, separators=(",", ": "), sort_keys=True
        )
        (file_url, fd) = self.core.call_success(
            "fs_mktemp", prefix="statistics_", suffix=".json", mode="w",
            on_disk=True
        )
        with fd:
            fd.write(stats)
        return file_url

    def on_bug_report_attachment_selected(self, attachment_id, *args):
        if attachment_id != 'stats':
            return

        self.core.call_all(
            "bug_report_update_attachment", attachment_id,
            {"file_url": _("Collecting statistics ...")},
            *args
        )

        node_uuid = self.core.call_success("config_get", "uuid")
        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._collect_stats, args=(node_uuid,)
        )
        promise = promise.then(self._write_stats_to_tmp_file)
        promise = promise.then(
            lambda file_url: self.core.call_all(
                "bug_report_update_attachment", attachment_id, {
                    'file_url': file_url,
                    'file_size': self.core.call_success(
                        'fs_getsize', file_url
                    ),
                }, *args
            )
        )
        promise.schedule()



import logging
import random

try:
    import gi
    GI_AVAILABLE = True
except (ImportError, ValueError):
    GI_AVAILABLE = False

try:
    GTK_AVAILABLE = False
    if GI_AVAILABLE:
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk
        GTK_AVAILABLE = True
except (ImportError, ValueError):
    pass

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps

from .... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_scan_buttons_popover_sources',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_doc_import',
                'defaults': ['paperwork_gtk.docimport'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_scan_buttons_popover',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageadd.source_popover'
                ],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GTK_AVAILABLE:
            return
        opt = self.core.call_success(
            "config_build_simple", "GUI", "last_import_uri",
            lambda: None
        )
        self.core.call_all("config_register", "last_import_uri", opt)

        self.core.call_all(
            "mainloop_schedule", self.core.call_all,
            "pageadd_sources_refresh"
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def pageadd_get_sources(self, out: list):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageadd", "source_box.glade"
        )
        source_long_txt = _("Import document or image file(s)")
        source_short_txt = _("Import file(s)")
        img = "insert-image-symbolic"

        widget_tree.get_object("source_image").set_from_icon_name(
            img, Gtk.IconSize.SMALL_TOOLBAR
        )
        widget_tree.get_object("source_name").set_text(source_long_txt)

        out.append(
            (
                widget_tree.get_object("source_selector"),
                source_short_txt, "import", self._on_import
            )
        )

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _on_import(self, doc_id, doc_url, source_id):
        LOGGER.info("Opening file chooser dialog")

        mimes = set()
        self.core.call_all("get_import_mime_types", mimes)
        mimes = list(mimes)
        mimes.sort()

        dialog = Gtk.FileChooserDialog(
            _("Select a file or a directory to import"),
            self.windows[-1],
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                # WORKAROUND(Jflesch): Use response ID 0 so the user
                # can select a folder.
                Gtk.STOCK_OPEN, 0
            )
        )
        dialog.set_select_multiple(True)
        dialog.set_local_only(False)

        filter_all = Gtk.FileFilter()
        filter_all.set_name(_("All supported file formats"))
        for (txt, mime) in mimes:
            filter_all.add_mime_type(mime)
        dialog.add_filter(filter_all)

        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("Any files"))
        file_filter.add_pattern("*.*")
        dialog.add_filter(file_filter)

        for (txt, mime) in mimes:
            file_filter = Gtk.FileFilter()
            file_filter.add_mime_type(mime)
            file_filter.set_name(txt)
            dialog.add_filter(file_filter)

        dialog.set_filter(filter_all)

        last_import_uri = self.core.call_success(
            "config_get", "last_import_uri"
        )
        if last_import_uri is not None:
            # this filename does not exist, but the parent directory
            # will be selected. A hack inspired by what firefox does.
            dialog.set_uri(last_import_uri + str(random.randint(0, 10000)))
        dialog.connect("response", self._on_dialog_response)

        self.core.call_all("pageadd_busy_add")

        dialog.show_all()

    def _on_dialog_response(self, dialog, response_id):
        self.core.call_all("pageadd_busy_remove")

        if (response_id != 0 and
                response_id != Gtk.ResponseType.ACCEPT and
                response_id != Gtk.ResponseType.OK and
                response_id != Gtk.ResponseType.YES and
                response_id != Gtk.ResponseType.APPLY):
            LOGGER.info("User canceled (response_id=%d)", response_id)
            dialog.destroy()
            return

        selected = dialog.get_uris()

        dialog.destroy()

        if len(selected) <= 0:
            notif = self.core.call_success(
                "get_notification_builder",
                _("Nothing to import")
            )
            notif.set_message(_("Nothing to import: nothing selected"))
            notif.set_icon("dialog-error")
            notif.show()
            return

        self.core.call_all("config_put", "last_import_uri", selected[0])

        self.core.call_all("gtk_doc_import", selected)



import gc
import logging
import os

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_delete_many"


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'action',
            'action_docs',
            'action_docs_delete',
            'chkdeps',
            'doc_action',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_selection',
                'defaults': ['paperwork_gtk.doc_selection'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_dialog_yes_no',
                'defaults': ['openpaperwork_gtk.dialogs.yes_no'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._delete)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def _delete(self, action, parameter):
        docs = set()
        self.core.call_all("doc_selection_get", docs)
        LOGGER.info("Asking confirmation before deleting %d doc", len(docs))
        msg = _('Are you sure you want to delete %d documents ?') % len(docs)

        if os.name == "nt":
            # On Windows, we have to be absolutely sure the PDF is actually
            # closed when we try to delete it because Windows s*cks pony d*cks
            # in h*ll.
            self.core.call_all("doc_close")
            # there is no "close()" method in Poppler
            gc.collect()

        self.core.call_all("gtk_show_dialog_yes_no", self, msg, docs)

    def on_dialog_yes_no_reply(self, parent, reply, *args, **kwargs):
        if parent is not self:
            return
        if not reply:
            return
        (docs,) = args

        self.core.call_all("doc_close")

        for (doc_id, doc_url) in docs:
            LOGGER.info("Deleting document %s", doc_id)
            self.core.call_all("storage_delete_doc_id", doc_id)

        self.core.call_all("search_update_document_list")

        self.core.call_success(
            "transaction_simple", [
                ('del', doc_id) for (doc_id, doc_url) in docs
            ]
        )



import gettext

import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'shortcuts',
            'shortcuts_page',
            'shortcuts_page_copy_text',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_page_copy_text',
                'defaults': ['paperwork_gtk.actions.page.copy_text'],
            },
            {
                'interface': 'app_shortcuts',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
        ]

    def init(self, core):
        super().init(core)

    def on_gtk_initialized(self):
        self.core.call_all(
            "app_shortcut_add",
            gettext.pgettext("keyboard shortcut categories", "Page"),
            gettext.pgettext(
                "keyboard shortcut names", "Copy selected text to clipboard"
            ),
            "<Control>c", "win.page_copy_text"
        )



import os
import tempfile

try:
    import fabulous.image
    # XXX(Jflesch): crappy workaround for an unmaintained library ...
    fabulous.image.basestring = str
    FABULOUS_AVAILABLE = True
except (ValueError, ImportError):
    FABULOUS_AVAILABLE = False

import openpaperwork_core


class FabulousRenderer(object):
    def __init__(self, plugin):
        self.plugin = plugin
        self.core = plugin.core
        self.parent = None

    def get_preview_output(
                self, doc_id, doc_url, terminal_size=(80, 25),
                page_idx=0
            ):
        w_split = int(terminal_size[0] / 3)

        parent = []
        if self.parent is not None:
            parent = self.parent.get_preview_output(
                doc_id, doc_url,
                (terminal_size[0] - w_split - 2, terminal_size[1]),
                page_idx
            )

        thumbnail = self.core.call_success(
            "thumbnail_get_page", doc_url, page_idx
        )
        if thumbnail is None:
            thumbnail = []
        else:
            thumbnail = self.plugin.img_render(thumbnail, w_split)

        if len(parent) < len(thumbnail):
            parent.extend([""] * (len(thumbnail) - len(parent)))
        elif len(parent) > len(thumbnail):
            parent = parent[:len(thumbnail)]

        out = [
            (i + " " + t)
            for (i, t) in zip(thumbnail, parent)
        ]

        return out

    def get_doc_output(self, doc_id, doc_url, terminal_size=(80, 25)):
        out = []
        if self.parent is not None:
            out = self.parent.get_doc_output(
                doc_id, doc_url, terminal_size
            )
        return out

    def get_page_output(
                self, doc_id, doc_url, page_nb, terminal_size=(80, 25)
            ):
        parent_out = []
        if self.parent is not None:
            parent_out = self.parent.get_page_output(
                doc_id, doc_url, page_nb, terminal_size
            )

        img_url = self.core.call_success("page_get_img_url", doc_url, page_nb)
        img = self.core.call_success("url_to_pillow", img_url)
        img = self.plugin.img_render(
            img, terminal_width=(terminal_size[0] - 1)
        )
        return [img_url] + list(img) + [""] + parent_out

    def get_doc_infos(self, doc_id, doc_url):
        out = {}
        if self.parent is not None:
            out = self.parent.get_doc_infos(doc_id, doc_url)
        return out

    def get_page_infos(self, doc_id, doc_url, page_nb):
        out = {}
        if self.parent is not None:
            out = self.parent.get_page_infos(doc_id, doc_url, page_nb)
        out['image'] = self.core.call_success(
            "page_get_img_url", doc_url, page_nb
        )
        return out


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def get_interfaces(self):
        return [
            'doc_renderer',
            'img_renderer',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'page_img',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
            {
                'interface': 'thumbnail',
                'defaults': ['paperwork_backend.model.thumbnail'],
            },
        ]

    def doc_renderer_get(self, out):
        if not FABULOUS_AVAILABLE:
            return
        r = FabulousRenderer(self)
        if len(out) > 0:
            r.parent = out[-1]
        out.append(r)

    def img_render(self, img, terminal_width=80):
        if not FABULOUS_AVAILABLE:
            return
        with tempfile.NamedTemporaryFile(
                    prefix='paperwork-shell', suffix='.jpeg',
                    delete=False
                ) as fd:
            img.save(fd, format="JPEG")
            img_file = fd.name
        try:
            img = fabulous.image.Image(img_file, width=terminal_width)
            img = img.reduce(img.convert())
        finally:
            os.unlink(img_file)
        return list(img)



import openpaperwork_core

from ... import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'shortcuts',
            'shortcuts_doc',
            'shortcuts_doc_copy_text',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_doc_new',
                'defaults': ['paperwork_gtk.actions.doc.new'],
            },
            {
                'interface': 'app_shortcuts',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
        ]

    def init(self, core):
        super().init(core)

    def on_gtk_initialized(self):
        self.core.call_all(
            "app_shortcut_add",
            _("Global"), _("Create new document"),
            "<Control>N", "win.doc_new"
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import sys

import openpaperwork_core
from openpaperwork_core.cmd.util import ask_confirmation

from .util import parse_page_list
from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "document_storage",
                "defaults": ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                "interface": "pages",
                "defaults": [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.thumbnail',
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'delete', help=_("Delete a document or a page")
        )
        p.add_argument(
            '--pages', '-p', type=str, required=False,
            help=_(
                "Pages to delete"
                " (single integer, range or comma-separated list,"
                " default: all pages)"
            )
        )
        p.add_argument(
            'doc_ids', nargs='*', default=[],
            help=_("Target documents")
        )

    def cmd_run(self, console, args):
        if args.command != 'delete':
            return None

        doc_ids = args.doc_ids

        for doc_id in doc_ids:
            if "/" in doc_id or "\\" in doc_id or ".." in doc_id:
                sys.stderr.write(f"Invalid doc_id: {doc_id}")
                sys.exit(2)

        pages = parse_page_list(args)

        del_doc_msg = _("Deleting document {doc_id} ...")
        del_page_msg = _("Deleting page {page_idx} of document {doc_id} ...")

        for doc_id in doc_ids:
            if pages is None:
                r = ask_confirmation(
                    console,
                    _("Delete document %s ?") % str(doc_id),
                    default_interactive='n',
                    default_non_interactive='y',
                )
            else:
                r = ask_confirmation(
                    _(
                        "Delete page(s)"
                        " {page_indexes} of document {doc_id} ?".format(
                            page_indexes=str([p + 1 for p in pages]),
                            doc_id=str(doc_id)
                        )
                    ),
                    default_interactive='n',
                    default_non_interactive='y',
                )
            if r != 'y':
                continue

            if pages is None:
                console.print(del_doc_msg.format(doc_id=doc_id))
                self.core.call_all("storage_delete_doc_id", doc_id)
            else:
                for page in pages:
                    doc_url = self.core.call_success(
                        "doc_id_to_url", doc_id
                    )
                    console.print(del_page_msg.format(
                        page_idx=(page + 1), doc_id=doc_id)
                    )
                    self.core.call_all("page_delete_by_url", doc_url, page)

        self.core.call_success(
            "transaction_simple", [
                # transaction_simple() will automatically replace the change
                # 'upd' by 'del' for the documents that don't exist anymore
                ("upd", doc_id) for doc_id in doc_ids
            ]
        )
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

        return True



import argparse
import json
import sys
import traceback

import rich.console
import rich.text

import openpaperwork_core
import openpaperwork_core.cmd

import paperwork_backend


# this import must be non-relative due to cx_freeze running this .py
# as an independant Python script
from paperwork_shell import _


DEFAULT_SHELL_PLUGINS = paperwork_backend.DEFAULT_PLUGINS + [
    "paperwork_backend.guesswork.cropping.libpillowfight",
    "paperwork_shell.cmd.chkworkdir",
    "paperwork_shell.cmd.delete",
    "paperwork_shell.cmd.edit",
    "paperwork_shell.cmd.export",
    "paperwork_shell.cmd.extra_text",
    "paperwork_shell.cmd.import",
    "paperwork_shell.cmd.label",
    "paperwork_shell.cmd.move",
    "paperwork_shell.cmd.ocr",
    "paperwork_shell.cmd.rename",
    "paperwork_shell.cmd.reset",
    "paperwork_shell.cmd.scan",
    "paperwork_shell.cmd.scanner",
    "paperwork_shell.cmd.search",
    "paperwork_shell.cmd.show",
    "paperwork_shell.cmd.sync",
    "paperwork_shell.display.docrendering.extra_text",
    "paperwork_shell.display.docrendering.img",
    "paperwork_shell.display.docrendering.labels",
    "paperwork_shell.display.docrendering.text",
    "paperwork_shell.l10n",
    "paperwork_shell.logs.print",
]

DEFAULT_CLI_PLUGINS = DEFAULT_SHELL_PLUGINS + [
    "paperwork_shell.cmd.about",
    "paperwork_shell.display.progress",
    "paperwork_shell.display.rich",
    "paperwork_shell.display.scan",
]
DEFAULT_JSON_PLUGINS = DEFAULT_SHELL_PLUGINS


class RichConsole:
    def __init__(self):
        self.console = rich.console.Console()

    def print(self, line):
        if isinstance(line, str):
            line = rich.text.Text(line)
        self.console.print(line)

    def input(self, prompt=""):
        return self.console.input(rich.text.Text(prompt + "\n"))


def main_main(in_args, application_name, default_plugins, interactive):
    # To load the plugins, we need first to load the configuration plugin
    # to get the list of plugins to load.
    # The configuration plugin may write traces using logging, so we better
    # enable and configure the plugin logs.print first.

    core = openpaperwork_core.Core()
    # plugin 'uncaught_exceptions' requires a mainloop plugin
    core.load('openpaperwork_core.mainloop.asyncio')
    core.load('paperwork_shell.logs.print')
    for module_name in paperwork_backend.DEFAULT_CONFIG_PLUGINS:
        core.load(module_name)
    core.init()

    core.call_all("config_load")
    core.call_all("config_load_plugins", application_name, default_plugins)

    core.call_all("init_logs", application_name, "warning")

    parser = argparse.ArgumentParser()
    cmd_parser = parser.add_subparsers(
        help=_('command'), dest='command', required=True
    )

    core.call_all("cmd_complete_argparse", cmd_parser)
    args = parser.parse_args(in_args)

    if interactive:
        console = RichConsole()
    else:
        console = openpaperwork_core.cmd.DummyConsole()

    core.call_all("cmd_set_console", console)
    if interactive:
        r = core.call_all("cmd_run", console=console, args=args)
    else:
        r = core.call_success("cmd_run", console=console, args=args)

    core.call_all("on_quit")
    return r


def json_main():
    try:
        r = main_main(
            sys.argv[1:], 'paperwork-json', DEFAULT_JSON_PLUGINS,
            interactive=False
        )
        print(json.dumps(
            r,
            indent=4,
            separators=(',', ': '),
            sort_keys=True
        ))
    except Exception as exc:
        stack = traceback.format_exc().splitlines()
        print(json.dumps(
            {
                "status": "error",
                "exception": str(type(exc)),
                "args": str(exc.args),
                "reason": str(exc),
                "stack": stack,
            },
            indent=4,
            separators=(',', ': '),
            sort_keys=True
        ))
        sys.exit(2)


def cli_main():
    main_main(
        sys.argv[1:], 'paperwork-cli', DEFAULT_CLI_PLUGINS, interactive=True
    )


if __name__ == "__main__":
    # Do not remove. Cx_freeze goes throught here
    if "paperwork-json" in sys.argv[0]:
        json_main()
    else:
        cli_main()



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class Scroller(object):
    def __init__(self, controller, widget):
        self.controller = controller
        self.widget = widget
        self.allocate_handler_id = None

    def goto(self):
        alloc = self.widget.get_allocation()
        if alloc.y < 0:
            self.allocate_handler_id = self.widget.connect(
                "size-allocate", self._goto_on_allocate
            )
        else:
            self._goto_on_allocate()

    def _goto_on_allocate(self, *args, **kwargs):
        alloc = self.widget.get_allocation()
        assert alloc.y >= 0
        self.controller.last_value = alloc.y
        self.controller.plugin.scroll.get_vadjustment().set_value(alloc.y)
        if self.allocate_handler_id is not None:
            self.widget.disconnect(self.allocate_handler_id)
            self.allocate_handler_id = None


class ScrollPageLockedController(BaseDocViewController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.last_upper = -1
        self.last_value = -1

    def enter(self):
        self.last_value = self.plugin.scroll.get_vadjustment().get_value()
        self.last_upper = self.plugin.scroll.get_vadjustment().get_upper()
        self.doc_goto_page(self.plugin.requested_page_idx)

    def doc_goto_page(self, page_idx):
        super().doc_goto_page(page_idx)
        for widget in self.plugin.page_layout.get_children():
            page = self.plugin.widget_to_page[widget]
            if page.page_idx != page_idx:
                continue
            Scroller(self, widget).goto()

    def on_page_size_obtained(self, page):
        super().on_page_size_obtained(page)
        self.doc_goto_page(self.plugin.requested_page_idx)

    def on_layout_size_allocate(self, layout):
        super().on_layout_size_allocate(layout)
        self.doc_goto_page(self.plugin.requested_page_idx)

    def docview_set_layout(self, name):
        super().docview_set_layout(name)
        self.doc_goto_page(self.plugin.requested_page_idx)

    def docview_set_zoom(self, zoom):
        super().docview_set_zoom(zoom)
        self.doc_goto_page(self.plugin.requested_page_idx)

    def docview_reload_page(self, page_idx):
        super().docview_reload_page(page_idx)
        self.doc_goto_page(self.plugin.requested_page_idx)

    def on_vscroll_changed(self, adj):
        super().on_vscroll_changed(adj)
        if adj.get_upper() == self.last_upper:
            return
        self.last_upper = adj.get_upper()
        self.doc_goto_page(self.plugin.requested_page_idx)

    def on_vscroll_value_changed(self, adj):
        super().on_vscroll_value_changed(adj)
        if adj.get_value() == self.last_value:
            return
        self.last_value = adj.get_value()
        self.plugin.docview_switch_controller('scroll', ScrollFreeController)


class ScrollFreeController(BaseDocViewController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.scroll_proportion = None

    def _upd_proportion(self):
        vadj = self.plugin.scroll.get_vadjustment()
        self.scroll_proportion = vadj.get_value() / vadj.get_upper()

    def enter(self):
        self._upd_proportion()

    def doc_goto_page(self, page_idx):
        self.plugin.docview_switch_controller(
            'scroll', ScrollPageLockedController
        )

    def _on_vscroll(self, adj):
        if self.scroll_proportion is None:
            return
        pos = self.scroll_proportion * adj.get_upper()
        self.scroll_proportion = None
        adj.set_value(pos)

    def docview_set_zoom(self, zoom):
        self._upd_proportion()

    def docview_set_layout(self, name):
        self._upd_proportion()

    def on_vscroll_changed(self, adj):
        super().on_vscroll_changed(adj)
        self._on_vscroll(adj)

    def on_vscroll_value_changed(self, adj):
        super().on_vscroll_value_changed(adj)
        self._on_vscroll(adj)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['scroll'] = ScrollPageLockedController(docview)



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import importlib
import logging

from .. import (_, PluginBase)


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        application = self.core.call_success("config_get_plugin_list_name")

        config_parser = parser.add_parser(
            'plugins', help=(_("Manage %s plugins") % application)
        )

        subparser = config_parser.add_subparsers(
            help=_("sub-command"), dest='subcommand', required=True
        )

        subparser.add_parser(
            'list', help=(
                _("Show plugins enabled for %s") % application
            )
        )

        p = subparser.add_parser(
            'add', help=(
                _("Add plugin in %s") % application
            )
        )
        p.add_argument('plugin_name')
        p.add_argument(
            '--no_auto', '-n', action="store_true",
            help=_("Do not correct dependencies automatically")
        )

        p = subparser.add_parser(
            'remove', help=(
                _("Remove plugin from %s") % application
            )
        )
        p.add_argument('plugin_name')
        p.add_argument(
            '--no_auto', '-n', action="store_true",
            help=_("Do not correct dependencies automatically")
        )

        subparser.add_parser(
            'reset', help=(_(
                "Clean up your mess by reseting the plugin list to its"
                " default value"
            ))
        )

        p = subparser.add_parser(
            'show', help=(
                _("Show information regarding a plugin (must be enabled)")
            )
        )
        p.add_argument('plugin_name')

    def cmd_run(self, console, args):
        if args.command != 'plugins':
            return None
        elif args.subcommand == "list":
            return self._cmd_list_plugins(console)
        elif args.subcommand == "add":
            return self._cmd_add_plugin(
                console, args.plugin_name, not args.no_auto
            )
        elif args.subcommand == "remove":
            return self._cmd_remove_plugin(
                console, args.plugin_name, not args.no_auto
            )
        elif args.subcommand == "reset":
            return self._cmd_reset_plugins(console)
        elif args.subcommand == "show":
            return self._cmd_show_plugin(console, args.plugin_name)
        else:
            return None

    def _cmd_add_plugin(
                self, console, plugin_name, auto=True, added=set(), save=True
            ):
        added.add(plugin_name)

        if auto:
            # Load the plugin, and check that all its dependencies are
            # satisfied.
            # Do not use the core to load it, otherwise we won't be able
            # to use 'call_success()' or 'call_all().
            try:
                module = importlib.import_module(plugin_name)
            except ModuleNotFoundError:
                console.print(
                    _("Error: Plugin '%s' not found") % plugin_name
                )
                return False
            plugin = module.Plugin()
            deps = plugin.get_deps()
            for dep in deps:
                try:
                    actives = self.core.get_by_interface(dep['interface'])
                    if len(actives) > 0:
                        continue
                except KeyError:
                    pass
                for default in dep['defaults']:
                    if default in added:
                        continue
                    console.print(
                        _(
                            "Adding plugin '{plugin_name}' to satisfy"
                            " dependency of '{other_plugin_name}' on interface"
                            " '{interface}'"
                        ).format(
                            plugin_name=default,
                            other_plugin_name=plugin_name,
                            interface=dep['interface']
                        )
                    )
                    r = self._cmd_add_plugin(
                        default, auto=True, added=added, save=False
                    )
                    if not r:
                        return r

        self.core.call_all("config_add_plugin", plugin_name)
        if save:
            self.core.call_all("config_save")
        console.print(_("Plugin {} added").format(plugin_name))
        return True

    def _cmd_remove_plugin(
            self, console, plugin_name, auto=True, removed=set(), save=True):
        removed.add(plugin_name)

        if auto:
            # look for plugins depending on this one
            # if they have no other plugin satisfying their dependency,
            # remove them too.
            for other_plugin in self.core.get_active_plugins():
                if other_plugin in removed:
                    continue
                deps = self.core.get_deps(other_plugin)
                for dep in deps:
                    actives = dep['actives']
                    actives = actives.difference(removed)
                    if len(actives) <= 0:
                        console.print(
                            _(
                                "Removing plugin '{plugin_name}' due to"
                                " missing dependency '{interface}'"
                            ).format(
                                plugin_name=other_plugin,
                                interface=dep['interface']
                            )
                        )
                        self._cmd_remove_plugin(
                            console, other_plugin, auto=auto, removed=removed,
                            save=False
                        )
                        break

        self.core.call_all("config_remove_plugin", plugin_name)
        if save:
            self.core.call_all("config_save")
        console.print(_("Plugin {} removed").format(plugin_name))
        return True

    def _cmd_list_plugins(self, console):
        plugins = self.core.call_success("config_list_plugins")
        console.print("  " + _("Active plugins:"))
        for plugin in plugins:
            console.print(plugin)
        return list(plugins)

    def _cmd_reset_plugins(self, console):
        self.core.call_success("config_reset_plugins")
        self.core.call_all("config_save")
        console.print("Plugin list reseted")
        return True

    def _print_columns(self, console, columns):
        out = ""
        for (column_size, string) in columns:
            out += "| "
            out += ("{:" + str(column_size) + "}").format(string)
        out = out[1:]
        console.print(out)

    def _get_printable_deps(
            self, plugin_name,
            parents_requirements=set(), depth=0, already_printed=set()):
        header = "|   " * (depth - 1)

        try:
            plugin = self.core.get_by_name(plugin_name)
        except KeyError:
            # Plugin not loaded --> can't get the info
            yield (
                header + "|-- " + plugin_name,
                "(not loaded)",
            )
            return

        str_plugin_name = plugin_name
        if plugin_name in already_printed:
            str_plugin_name = plugin_name + " (dup)"

        interfaces = plugin.get_interfaces()
        interfaces = [i for i in interfaces if i in parents_requirements]
        if len(interfaces) <= 0:
            interfaces = [""]

        for (idx, interface) in enumerate(interfaces):
            if idx == 0:
                yield (
                    header + "|-- " + str_plugin_name
                    if depth > 0 else
                    str_plugin_name,
                    interface,
                )
            else:
                yield (
                    header + "|   |",
                    interface,
                )

        if plugin_name in already_printed:
            return []
        already_printed.add(plugin_name)

        deps = plugin.get_deps()

        requirements = {d['interface'] for d in deps}
        requirements.update(parents_requirements)

        plugin_names = set()
        for (idx, dep) in enumerate(deps):
            plugins = self.core.get_by_interface(dep['interface'])
            plugin_names.update({plugin.__module__ for plugin in plugins})
            plugin_names.update(dep['defaults'])

        plugin_names = list(plugin_names)
        plugin_names.sort()
        for (idx, plugin_name) in enumerate(plugin_names):
            for line in self._get_printable_deps(
                    plugin_name, requirements, depth + 1, already_printed):
                yield line

    def _cmd_show_plugin(self, console, plugin_name):
        try:
            plugin = self.core.get_by_name(plugin_name)
        except KeyError:
            console.print(_("Plugin '%s' not enabled.") % plugin_name)
            return {}

        console.print(_("Plugin '%s':") % plugin_name)
        console.print("* " + _("Implements:"))
        for interf in plugin.get_interfaces():
            console.print("  + " + interf)
        console.print("* " + _("Depends on:"))
        for dep in plugin.get_deps():
            console.print("  + " + dep['interface'])
            for default in dep['defaults']:
                console.print("    - " + (_("suggested: %s") % default))

        console.print("")

        deps = list(self._get_printable_deps(plugin_name))

        column_headers = (
            _("Plugin name"),
            _("Interface"),
        )

        column_sizes = [len(c) + 1 for c in column_headers]
        for d in deps:
            for (idx, column_value) in enumerate(d):
                column_sizes[idx] = max(
                    column_sizes[idx], len(column_value) + 1
                )
        total = sum(column_sizes) + (2 * len(column_sizes))

        self._print_columns(
            console,
            (
                (column_sizes[0], _("Plugin name")),
                (column_sizes[1], _("Interface")),
            )
        )
        console.print("-" * total)
        for d in deps:
            self._print_columns(
                console, [
                    (column_sizes[idx], column_value)
                    for (idx, column_value) in enumerate(d)
                ]
            )

        return {
            'interface': plugin.get_interfaces(),
            'deps': plugin.get_deps()
        }



import gc
import logging

import pympler.tracker

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.tracker = None

    def get_interfaces(self):
        return ['memleak_detector']

    def on_memleak_track_start(self):
        self.tracker = None
        LOGGER.info("Garbage collecting ...")
        gc.collect()
        LOGGER.info("Starting memory leaks tracking ...")
        self.tracker = pympler.tracker.SummaryTracker()

    def on_memleak_track_stop(self):
        if self.tracker is not None:
            LOGGER.info("Garbage collecting ...")
            gc.collect()
            LOGGER.info("Stopping memory leaks tracking ...")
            self.tracker.print_diff()
        self.tracker = None
